#!/usr/bin/env python

from Transport 		import ProcessClient
from Transport		import NewsPostMixin

import sys

class Writer(ProcessClient,NewsPostMixin):
	
	def __init__(self,argv):
		ProcessClient.__init__(self,argv)
		NewsPostMixin.__init__(self)

		self.pollrate	= self.getDeltaTime('rate',60)
		self.message	= self.get('message')

	def run(self):

		while self.running:

			self.newsPoster.postText(self.message)
			self.log.info('Posted message')
			self.wait(self.pollrate)

if __name__ == '__main__':
	Writer(sys.argv).run()
	



#!/usr/bin/env python

##########################################################
#
#   Example 2 - post a news message 
#
#   Post a message to the news group 
#
#   2010-08-16  Todd Valentic
#               Initial implementation
#
##########################################################

from Transport import ProcessClient
from Transport import NewsPostMixin

import sys

class PostExample(ProcessClient,NewsPostMixin):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)
        NewsPostMixin.__init__(self)

        self.rate    = self.getDeltaTime('rate','1:00')
        self.message = self.get('message')

        if self.message is None:
            self.abort('You need to specify a message')

        self.log.info('Rate: %s' % self.rate)

    def run(self):

        while self.wait(self.rate,sync=True):
            self.newsPoster.postText(self.message)
            self.log.info('Posted message')

if __name__ == '__main__':
    PostExample(sys.argv).run()




#!/usr/bin/env python
"""News Tool"""

# pylint: disable=too-many-lines

###########################################################################
#
#   News Tool
#
#   A collection of classes to support pushing and pulling of data from
#   newsgroups using NNTP. The object heirarchy looks like this:
#
#       NewsTool
#          |
#          |---- NewsPoster
#          |         |
#          |          ---- NewsControl
#          |
#           ---- NewsPoller
#
#   Revision History:
#
#   1.0.0   1999-??-??  Todd Valentic
#           Initial version
#
#   1.0.1   1999-09-17  Todd Valentic
#           Fix bug in NewsPoller that prevented new messages from being
#               detected. The comparison of the message numbers was done
#               as strings, which failed in one particular case ("97">"258")
#               but seemed to work in other situations. I forced the
#               comparison to be done on ints. The index numbers should
#               really be stored as ints and not strings - to be cleaned up
#               in the future.
#
#   1.0.2   1999-12-03  Todd Valentic
#           Modified the NewsPoller callback's to include the header from
#               the extracted message (used primarily to extract the date
#               of posting without having to interpret the file contents.)
#           Cleaned up the polling loop syntax.
#
#   1.0.3   2000-01-20  Todd Valentic
#           Added try..except blocks around processing calls in NewsPoller
#           Added logging function instead of just going to stdout.
#           Added markRead function to NewsPoller.
#
#   1.0.4   2000-01-22  Todd Valentic
#           Added openServer() method the NewsToolBase for connections to
#           the NNTP server. The reason for this was to include a
#           "mode reader" command if necessary when connecting. If a machine
#           is defined as a feeder for a downstream server, it will not be
#           allowed reading access. The innd program adds the "mode" command
#           to allow a server to force reader access. Since all the commands
#           in this library are for readers, we go ahead and try to force
#           the access mode for the connection.
#
#   1.0.5   2000-01-26  Todd Valentic
#           Modifications to NewsPoller:
#               - Added ability to pass raw article to processing function
#               - Removed the processScript option
#               - Unpack the MIME message only if necessary
#               - Removed the 'body.txt' check
#               - Added debug flag to be used when debugging processing
#           Modifications to NewsPoster:
#               - Added ability to post a non-MIME text only body.
#
#   1.0.6   2000-01-31  Todd Valentic
#           Added single_shot setting to NewsPoller. This flag indicates that
#               only a single message should be processed even though more
#               then one may be available.
#
#   1.0.7   2000-02-15  Todd Valentic
#           Updates to NewsPoller:
#               - Split processing from polling loop to processArticle.
#                 This is to allow other methods process articles in a
#                 consistent manner.
#               - Added processPast() to process articles in a group
#                 prior to the current time. No checks are made to see
#                 if the articles had previously been processed nor is
#                 the lastreadfile updated.
#           Updates to NewsToolBase:
#               - Added time() to query the newsserver for it's
#                 current time and return it as a Unix time.
#               - Added optional name parameter to openServer to allow
#                 setting the server name at the same time.
#
#   1.0.8   2000-04-24  Todd Valentic
#           Updates to NewsPoller:
#               - Added Retry exception
#               - Moved processArticle guts to callProcessing
#               - processArticle now retries the processing if the
#                 exception is raised. It will continue doing this
#                 until the processing functions run without error.
#                 This should handle the cases when we are trying to send
#                 data to a client such as SPARC and we suffer network
#                 outages.
#               - Added enable functionality.
#
#   1.0.9   2000-04-29  Todd Valentic
#           Moved groupExsits() method to NewsToolBase from NewsControl.
#
#   1.0.10  2000-06-22  Todd Valentic
#           Added try..except blocks around the server.quit() calls.
#               It seems that we were having problems posting to SPARC
#               in the processing calls (leading to long timeouts). In the
#               mean time, the connection to the news server had been
#               dropped. The call then to server.quit raised an exception.
#               I'm not sure we actually need to do a quit() - it might
#               be better to have a persistent server object sitting
#               around that reopens a connection when needed.
#
#   1.0.11  2001-08-11  Todd Valentic
#           Added processText function to newsPoller to act like the
#                 other processing methods (func,raw). It separates the
#                 headers from the body.
#
#   1.0.12  2001-09-06  Todd Valentic
#           Changes set_newsgroup() to handle a list of group names as well
#                 as just an individual string to accommodate cross-posting.
#
#   1.0.13  2001-11-28  Todd Valentic
#           Chnaged header in postText to include data field like the
#                 normal post() method uses.
#
#   1.0.14  2002-01-24  Todd Valentic
#           Changed processFunc -> processFiles.
#
#   1.0.15  2002-04-11  Todd Valetic
#           Changed the body creation in NewsPoller processText() to
#                 join lines with a \n.
#
#   1.0.16  2002-08-25  Todd Valentic
#           Use readermode in NNTP() to set reader mode (see discussion
#                 in 1.0.4).
#
#   1.0.17  2002-10-10  Todd Valentic
#           Added check in callProcessing() for bad message id.
#
#   1.0.18  2003-02-12  Todd Valentic
#           Made catchup parameter more flexible for polling:
#                   catchup = 0: do nothing
#                   catchup = 1: mark all messages read
#                   catchup < 0: mark all messages except last count as read
#                                (but not if we have already seen them).
#
#   1.0.19  2003-04-10  Todd Valentic
#           Start using mx.DateTime module.
#           Cleaned up formatting of comments.
#           Removed setComment in NewsPoster - replaced with parameter to post
#           Time added to subject lines now in ARPA format.
#           Use string methods instead of string module.
#           There was a time that article numbers were missing the trailing
#             '>' when queried from INN. This doesn't seem to happen anymore.
#             Removing check code:
#               if article_number[-1]!='>':
#                 article_number=article_number+'>'
#           Added listArticles and listGroups methods
#           Removed setOutputName, setPrefix in NewsPoller (not used anymore)
#           Renamed NewsToolBase to NewsTool
#           Added stop() callback to NewsPoller to stop processing when
#               dealing with a large number of backlogged articles.
#           Added 'X-Transport-Date' header and parameters to postFiles() and
#               postText(). This allows a client to specify a date to later
#               be used for special purposes. For example, a client might set
#               date to match the time at which a data file is generated, not
#               the time at which the file is posted. An archiving program
#               can then use this date file to name the file correctly. This
#               used to be handled by coding the date in the Subject header.
#
#   1.0.20  2003-04-18  Todd Valentic
#           Fixed some bugs due in NewsControl object that were introduced
#               in code cleanup.
#
#   1.0.21  2003-06-10  Todd Valentic
#           Somewhere along the line, a bug was introduced where extra
#               characters were added to the end of unpacked files! This
#               was a problem with the old mimemsg interface. I replaced
#               it with the much simpler standard library email.
#
#   2.0.0   2003-06-30  Todd Valentic
#           Changed the signature for the callback functions. They now
#               provide an email.Message object as the first parameter
#               followed by the specific callback parameters (list of
#               filenames, etc). The subject line is no longer directly
#               supplied. It can easily be obtained from the message
#               headers.
#           Added the messageDate() function to the utility methods.
#           Renamed SaveFiles() -> saveFiles(). It now takes a Message.
#
#   2.0.1   2003-08-21  Todd Valentic
#           Fixed bug in processPast - didn't handle list of newsgroups
#               returned by listArticles properly.
#
#   2.0.2   2004-02-13  Todd Valentic
#           Pass user-defined headers along when posting.
#
#   2.0.3   2004-02-17  Todd Valentic
#           Record message time/size info when reading.
#
#   2.0.4   2004-02-26  Todd Valentic
#           Check if multipart in saveFiles. For non-MIME messages, save
#               the body insead.
#
#   2.0.5   2004-09-11  Todd Valentic
#           Added asConfig() utility method.
#
#   2.0.6   2004-12-27  Todd Valentic
#           Convert from mx.DateTime to datetime
#           Use dateutil.parser for date parsing
#           Change time() method to return server time in UTC
#           Change time method name to datetime
#
#   2.0.7   2005-01-10  Todd Valentic
#           Make sure all times are UTC.
#
#   2.0.8   2005-02-08  Todd Valentic
#           Fix default logging
#           Allow passing a single filename into post.
#           Add postFiles() to compliment other post* methods. This
#               is just an alias for post.
#
#   2.0.9   2005-03-08  Todd Valentic
#           Handle exception in callProcessing
#
#   2.0.10  2005-03-25  Todd Valentic
#           Added path prefix and creation to saveFiles
#
#   2.0.11  2005-04-04  Todd Valentic
#           Use get_content_* methods instead of deprectated verions.
#
#   2.0.12  2005-05-09  Todd Valentic
#           Reorganized poll() internals. Added getNextMessage() for
#               use by other routines (i.e. the SyncPoller component).
#           Added ability to set port when specifying server.
#           Added mark_message_read()
#
#   2.0.13  2005-06-13  Todd Valentic
#           Added set/add/clearExtraHeaders() to NewsPoster.
#           Added getArticle - fixed bug in processPast.
#
#   2.0.14  2005-07-29  Todd Valentic
#           Cleaned up listGroups to use fnmatch.filter.
#
#   2.0.15  2005-12-21  Todd Valentic
#           Catchup lastid if it is less then the first available
#               message id.
#           Fixed bug in Poller run() routine. Bad return value
#               in the case the server was unavailable. Found by
#               Andrew Young.
#
#   2.0.16  2006-02-03  Todd Valentic
#           Code cleanup. Removed deprecated verbose entries.
#
#   2.0.17  2006-02-07  Todd Valentic
#           Make sure that messageDate() has a timezone. Default
#               to UTC if not specified.
#
#   2.0.18  2006-07-18  Todd Valentic
#           Extra headers were not being written out in postText
#           Remove timestamp on subject line
#
#   2.0.19  2006-10-14  Todd Valentic
#           Make headers parameter to post* methods a dict.
#
#   2.0.20  2007-07-17  Todd Valentic
#           Added listFiles() and write parameter to saveFiles().
#
#   2.0.21  2008-01-06  Todd Valentic
#           Fixed bug in saveFiles where path was ignored on text
#               only messages.
#
#   2.0.22  2008-06-07  Todd Valentic
#           Fixed off by catchup off-by-one error
#
#   2.0.23  2008-08-15  Todd Valentic
#           Ignore path in saveFiles unless it is set (fix in 2.0.21
#               added ./ to all files otherwise, which is unexpected
#               by existing programs).
#
#   2.0.24  2008-08-19  Todd Valentic
#           Only use basename for attached filenames in saveFiles().
#
#   2.0.25  2008-11-13  Todd Valentic
#           Use SafeConfigParser.
#
#   2.0.26  2009-05-04  Todd Valentic
#           Fix bug in setting message header for X-Transport-ArticleNumber.
#               It an email object already has a header set, setting it
#               again does not overwrite the old definition, instead it
#               appends to the end of the list. Make sure to del it first.
#
#   2.0.27  2009-05-11  Todd Valentic
#           Use email instead of MimeWriter.
#           Code cleanup.
#           Use header array.
#
#   2.0.28  2009-08-24  Todd Valentic
#           Set socket timeout on server.
#
#   2.0.29  2009-11-26  Todd Valentic
#           Added getContents()
#
#   2.0.30  2010-05-07  Todd Valentic
#           Add check in getNextMessage() for situations where message
#               count has gone backwards. This can happen when moving
#               to a new news server.
#
#   2.0.31  2014-01-17  Todd Valentic
#           Add exclude to listGroups()
#
#   2.0.32  2016-10-31  Todd Valentic
#           Fix iteration bug introducted in listGroups(exclude)
#
#   2.0.33  2020-01-09  Todd Valentic
#           Add last_read_prefix to NewsPoller
#
#   2016-12-23  Todd Valentic
#               Use datetime.timezone.utc for UTC
#
#   2022-10-06  Todd Valentic
#               Reorder imports
#               Use sapphire.Parser
#               Use pathlib
#               Python3 changes:
#                   - Update expceptions
#                   - Update imports
#                   - Use bytes instead of string
#                   - article() return changed to (response, info)
#               Snake names:
#                   getContents -> get_contents
#                   saveFiles -> save_files
#                   listFiles -> list_files
#                   messageDate -> message_date
#                   asConfig -> as_config
#
#               NewsTool
#                   setServer -> set_server
#                   setGroup -> set_newsgroup
#                   setLog - > set_log
#                   setTimeout - set_timeout
#                   openServer -> open_server
#                   groupExists -> group_exists
#                   datetime - > get_datetime
#                   listArticles -> list_articles
#                   listGroups -> list_newsgroups
#
#   2023-07-26  Todd Valentic
#               Support pathlib in poster
#
#   2023-08-07  Todd Valentic
#               Set policy for longer line length (150 chars). We were
#                   failing to create new groups with Control messages
#                   when the Control header exceeded the default 77 
#                   characters. The Python splitting altogrithm worked,
#                   but innd does not appear to support multiline headers.
#
#   2023-08-23  Todd Valentic
#               Add ability to set path for NewsPoller last read file
#               Create last read file path if necessary
#
###########################################################################

import collections
import datetime
import email
import logging
import mimetypes
import nntplib
import pathlib
import time

from email import encoders
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from fnmatch import fnmatch

from dateutil import parser
from datatransport.utilities import datefunc, make_path
import sapphire_config as sapphire

utc = datetime.timezone.utc

###### Exception Class ###################################################


class ProcessError(Exception):
    """Raised on processing errors"""


class ProcessRetry(Exception):
    """Raised to request retry on processing error"""


##########################################################################
#
#   Utility Methods
#
##########################################################################


def get_contents(message):
    """Return body and/or contents of each attachment in a list"""

    results = []

    if not message.is_multipart():
        results.append(message.get_payload())
    else:
        for part in message.walk():
            if part.get_content_maintype() == "multipart":
                continue
            results.append(part.get_payload(decode=True))

    return results


def save_files(message, default="body.txt", path=".", write=True):
    """Save body or attached files"""

    path = pathlib.Path(path)

    if not message.is_multipart():
        if write:
            body = message.get_payload()
            filename = path / default
            make_path(filename)
            filename.write_text(body)
        return [filename]

    filenames = []

    counter = 1
    for part in message.walk():
        if part.get_content_maintype() == "multipart":
            continue
        filename = part.get_filename()
        if not filename:
            ext = mimetypes.guess_extension(part.get_content_type())
            if not ext:
                ext = ".bin"
            filename = f"part-{counter:03}{ext}"
        counter += 1

        filename = path / pathlib.Path(filename).name

        if write:
            make_path(filename)
            filename.write_bytes(part.get_payload(decode=True))

        filenames.append(filename)

    return filenames


def list_files(*pos, **kw):
    """Return list of attached files"""

    kw["write"] = False
    return save_files(*pos, **kw)


def message_date(message):
    """Return the message timestamp"""

    headers = ["X-Transport-Date", "NNTP-Posting-Date", "Date"]

    dateparser = parser.parser()

    for header in headers:
        if header in message:
            timestamp = dateparser.parse(message[header])
            if timestamp.tzinfo is None:
                timestamp = timestamp.replace(tzinfo=utc)
            return timestamp

    return datetime.datetime.now(utc)


def as_config(message):
    """Return the message content as a sapphire.Parser object"""

    config = sapphire.Parser()
    config.read_string(message.get_payload())

    return config


##########################################################################
#
#   Base Class
#
##########################################################################


class NewsTool:
    """NewsTool base class"""

    def __init__(self):
        self.set_log(logging)
        self.set_server("localhost")
        self.set_newsgroup("test")
        self.set_timeout(60)

    # Configuration variables ---------------------------------------------

    def set_server(self, host, port=119):
        """Set newsserver parameters"""

        self.server_host = host
        self.server_port = port

    def set_newsgroup(self, newsgroup):
        """Set newsgroups header"""

        if isinstance(newsgroup, str):
            self.newsgroup_header = newsgroup
        else:
            self.newsgroup_header = ",".join(newsgroup)

    def set_log(self, func):
        """Set logger"""

        self.log = func

    def set_timeout(self, secs):
        """Set timeout"""

        self.timeout = secs

    # Services -----------------------------------------------------------

    def open_server(self, host=None, port=119):
        """Initialize instance"""

        if host:
            self.set_server(host, port)

        return nntplib.NNTP(self.server_host, 
                            port=self.server_port, 
                            readermode=True, 
                            timeout=self.timeout)

    def has_newsgroup(self, newsgroup=None):
        """Check if newsgroup exists"""

        if not newsgroup:
            newsgroup = self.newsgroup_header

        try:
            server = self.open_server()
            server.group(newsgroup)
            server.quit()
        except nntplib.NNTPTemporaryError as err:
            code = str(err).split()[0]
            if code == "411":
                return False
            raise err

        return True

    def get_datetime(self, server=None):
        """Get time from the server"""

        if server is None:
            server = self.open_server()

        response = server.date()
        datestr = response[0].split()[1]  # 111 YYYYMMDDhhmmss

        return datefunc.strptime(datestr, "%Y%m%d%H%M%S", tzinfo=utc)

    def list_articles(self, offset, newsgroups=None):
        """
        List articles during offset (which is type timedelta)

        offset is a timedelta

        newsgroup can be: None (use the header member)
                          str - single group name
                          list - list of group names
        """

        server = self.open_server()

        start = self.get_datetime(server) - offset

        if newsgroups is None:
            newsgroups = self.newsgroup_header.split(",")

        if isinstance(newsgroups, str):
            newsgroups = [newsgroups]

        articles = {}

        for newsgroup in newsgroups:
            msgs = server.newnews(newsgroup, start)[1]
            articles[newsgroup] = msgs

        return articles

    def list_newsgroups(self, pattern="transport.*", exclude=None):
        """List newsgroups matching pattern"""

        if exclude is None:
            exclude = []
        elif exclude and isinstance(exclude, str):
            exclude = [exclude]

        _response, newsgroups = self.open_server().list()
        newsgroups = [ng for ng in newsgroups if fnmatch(ng[0], pattern)]

        for entry in exclude:
            newsgroups = [ng for ng in newsgroups if not fnmatch(ng[0], entry)]

        results = {}

        for entry in newsgroups:
            minmsg = int(entry[2])
            maxmsg = int(entry[1])
            if maxmsg < minmsg:
                results[entry[0]] = (minmsg, maxmsg, 0)
            else:
                results[entry[0]] = (minmsg, maxmsg, maxmsg - minmsg + 1)

        return results


#########################################################################
#
#   News Poster
#
#########################################################################


class NewsPoster(NewsTool):
    """News Poster"""

    def __init__(self):
        self.clear_headers()
        NewsTool.__init__(self)

        self.set_subject("Unknown")
        self.set_from("transport@datatransport.org")
        self.set_enable(True)

    def set_newsgroup(self, newsgroup):
        """Set newsgroup header"""

        NewsTool.set_newsgroup(self, newsgroup)
        if newsgroup:
            self.set_header("Newsgroups", newsgroup)

    def set_subject(self, value):
        """Set subject header"""

        self.set_header("Subject", value)

    def set_from(self, value):
        """Set from header"""

        self.set_header("From", value)

    def set_enable(self, flag):
        """Set the enable flag"""

        self.enabled = flag

    def set_header(self, key, value):
        """Set header value"""

        self.headers[key] = value

    def clear_headers(self):
        """Clear all headers"""

        self.headers = {}

    def add_headers(self, msg, date=None, extra=None):
        """Add headers to message"""

        if date:
            msg["X-Transport-Date"] = str(date)

        for key, value in self.headers.items():
            msg[key] = value

        if extra:
            for key, value in extra.items():
                msg[key] = value

    def add_file(self, msg, filename):
        """Add file attachment to message"""

        filename = pathlib.Path(filename)

        ctype, encoding = mimetypes.guess_type(filename)
        if ctype is None or encoding is not None:
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)
        if maintype == "text":
            contents = filename.read_text("UTF-8")
            part = MIMEText(contents, _subtype=subtype)
        elif maintype == "image":
            contents = filename.read_bytes()
            part = MIMEImage(contents, _subtype=subtype)
        elif maintype == "audio":
            contents = filename.read_bytes()
            part = MIMEAudio(contents, _subtype=subtype)
        else:
            contents = filename.read_bytes()
            part = MIMEBase(maintype, subtype)
            part.set_payload(contents)
            encoders.encode_base64(part)

        basename = filename.name
        part.add_header("Content-Disposition", "attachment", filename=basename)

        msg.attach(part)

    def post(self, filenames=None, comment=None, date=None, headers=None):
        """Post files to newsgroup"""

        if filenames is not None:
            if isinstance(filenames, str):
                filenames = [filenames]
            elif not isinstance(filenames, collections.abc.Iterable): 
                filenames = [filenames]

        if not filenames:
            msg = MIMEText(comment)
        else:
            msg = MIMEMultipart()
            msg.preable = comment

            for filename in filenames:
                self.add_file(msg, filename)

        self.add_headers(msg, date=date, extra=headers)

        return self.post_raw(msg)

    def post_files(self, *pos, **kw):
        """Post files to newsgroup (alias for post)"""

        return self.post(*pos, **kw)

    def post_text(self, text, *pos, **kw):
        """Post text message to newsgroup"""

        return self.post(comment=text, *pos, **kw)

    def post_raw(self, msg):
        """Post message to newsgroup if enabled"""

        if self.enabled:
            policy = msg.policy.clone(max_line_length=150)
            return self.open_server().post(msg.as_bytes(policy=policy))

        return None


##########################################################################
#
#   News Control
#
##########################################################################


class NewsControl(NewsPoster):
    """Post control messages to news server"""

    def __init__(self):
        NewsPoster.__init__(self)

        self.set_newsgroup("control")

    def post_command(self, cmd, body=None):
        """Post control message"""

        if not body:
            body = cmd

        self.set_header("Control", cmd)
        self.set_header("Approved", self.headers["From"])
        return self.post(comment=body)

    def newgroup(self, newsgroup):
        """Create a newsgroup"""

        body = f"For your newsgroups file:\n{newsgroup} {newsgroup}"
        return self.post_command(f"newgroup {newsgroup}", body)

    def rmgroup(self, newsgroup):
        """Remove a newsgroup"""

        return self.post_command(f"rmgroup {newsgroup}")

    def cancel_message(self, newsgroup, message_id):
        """Cancel a message in a newsgroup"""

        self.set_newsgroup(newsgroup)
        return self.post_command(f"cancel {message_id}")

    def cancel_newsgroup(self, newsgroup):
        """Cancel all of the messages in a newsgroup"""

        server = self.open_server()
        _response, _count, first, last, _name = server.group(newsgroup)
        _response, subject = server.xhdr("Message-ID", f"{first}-{last}")
        server.quit()

        for _article_number, message_id in subject:
            self.cancel_message(newsgroup, message_id)


##########################################################################
#
#   News Poller
#
##########################################################################


class NewsPoller(NewsTool):
    """News Poller"""

    def __init__(self):
        NewsTool.__init__(self)

        self.set_callback(self.default_callback)
        self.set_debug(False)
        self.set_single_shot(False)
        self.set_retry_wait(60)
        self.set_stop_func(self.default_stop)
        self.set_last_read_prefix("")
        self.set_last_read_path(".")

    # Configuration variables ---------------------------------------------

    def set_callback(self, func):
        """Set callback processing function"""

        self.callback = func

    def set_debug(self, flag):
        """Set debug flag"""
        self.debug = flag

    def set_single_shot(self, flag):
        """Set single shot flag"""
        self.single_shot = flag

    def set_retry_wait(self, secs):
        """Set retry wait time"""
        self.retry_wait = secs

    def set_stop_func(self, func):
        """Loop exit condition"""
        self.stop = func

    def set_last_read_prefix(self, prefix):
        """Prefix for last read tracking file"""
        if prefix:
            self.last_read_prefix = prefix + "-"
        else:
            self.last_read_prefix = ""

    def set_last_read_path(self, path):
        self.last_read_path = pathlib.Path(path)

    # Methods -------------------------------------------------------------

    def default_stop(self):
        """Place holder - clients can override"""
        return False

    def default_callback(self):
        """Place holder - set by clients"""
        return

    def last_read_filename(self):
        """Return the filename of the last_read tracking file"""
        filename = self.last_read_prefix + self.newsgroup_header
        return self.last_read_path.joinpath(filename)

    def save_last_read(self, article_num):
        """Update the last read tracking file"""
        output = self.last_read_filename()
        output.parent.mkdir(parents=True, exist_ok=True)
        self.last_read_filename().write_text(str(article_num), "UTF-8")

    def load_last_read(self):
        """Read the last read tracking file"""
        article_num = self.last_read_filename().read_text("UTF-8")
        return int(article_num.strip())

    def mark_read(self, msgcount=1, reset=False):
        """Mark newsgroup as read"""

        if msgcount == 0:
            return

        server = self.open_server()
        response = server.group(self.newsgroup_header)
        server.quit()

        first = int(response[2])
        last = int(response[3])

        try:
            lastid = self.load_last_read()
        except (FileNotFoundError, ValueError, UnicodeDecodeError):
            lastid = first - 1

        if msgcount == 1:
            newlast = last
        elif reset:
            newlast = max(0, last + msgcount)
        else:
            newlast = max(lastid, last + msgcount)

        self.log.debug(
            f"Marking read ({msgcount}): "
            f"old watermark={lastid}, new watermark={newlast}"
        )

        self.save_last_read(newlast)

    def call_processing(self, message):
        """Call the processing callback function"""

        article_number = message["X-Transport-ArticleNumber"]

        self.log.debug(f"  Processing message {article_number}")

        try:
            self.callback(message)
        except ProcessRetry:
            raise  # propogate up
        except BaseException as err:  # pylint: disable=broad-except
            self.log.error(f"Problem in the callback function: {err}")

            if self.debug:
                raise err

    def process_article(self, message):
        """Process the article"""

        while not self.stop():

            try:
                self.call_processing(message)
                self.mark_message_read(message)
                return
            except ProcessRetry:
                time.sleep(self.retry_wait)

            # FIX ME - should pass in/use wait

    def mark_message_read(self, message):
        """Mark this message has read"""

        self.save_last_read(message["X-Transport-ArticleNumber"])

    def process_past(self, offset):
        """Process old messages"""

        self.log.debug("Processing past articles:")

        server = self.open_server()

        for newsgroup, articles in self.list_articles(offset=offset).items():
            self.log.debug(f"  {newsgroup}: {len(articles)} articles available")

            for article_num in articles:

                if self.stop():
                    break

                try:
                    message = self.get_article(server, article_num)
                except:  # pylint: disable=bare-except
                    continue

                # TBD - Better except handling

                self.process_article(message)

    def get_article(self, server, article_num):
        """Retrieve message from the newsgroup"""

        _response, info = server.article(article_num)
        content = b"\n".join(info.lines)
        message = email.message_from_bytes(content)

        del message["X-Transport-ArticleNumber"]
        message["X-Transport-ArticleNumber"] = article_num

        return message

    def get_next_message(self):
        """Get the next message from the newsgroup"""

        try:
            server = self.open_server()
            _resp, count, low, high, name = server.group(self.newsgroup_header)
        except nntplib.NNTPError as e:
            self.log.debug("Failed to get message numbers: %s", e)
            return None
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.log.debug("Failed to get message numbers: %s", e)
            return None

        self.log.debug(f"Group {name} has {count} articles from {low} to {high}")

        try:
            lastid = self.load_last_read()
            self.log.debug(f"  last message read was {lastid}")
        except:  # pylint: disable=bare-except
            lastid = low - 1
            self.log.debug("  no messages read yet")

        if low > high:
            self.log.debug("  no messages on server")
            return None

        if lastid < low - 1 or lastid > high:
            self.log.debug("  catching up to available messages")
            lastid = low - 1

        nextid = lastid + 1

        if nextid > high:
            self.log.debug("  no new messages")
            return None

        article_number = str(nextid)

        self.log.debug("  retrieving  message {article_number}")

        try:
            message = self.get_article(server, article_number)
        except nntplib.NNTPTemporaryError as e:
            code = int(str(e).split()[0])
            if code >= 400:
                self.log.error("Problem retrieving {article_num} (code={code})")
                self.save_last_read(article_number)
            return None
        except:  # pylint: disable=bare-except
            self.log.exception("Problem parsing message body")
            self.save_last_read(article_number)
            return None

        return message

    def poll(self):
        """Poll newsgroup for new messages"""

        while not self.stop():

            message = self.get_next_message()

            if message is None:
                break

            self.process_article(message)

            if self.single_shot:
                break

        self.log.debug("End of polling cycle")



#!/usr/bin/env python

##################################################
#
#   Uptime Service
#
#   2010-08-19  Todd Valentic
#               Initial implementation
#
##################################################

from Transport  import ProcessClient
from Transport  import XMLRPCServerMixin

import sys

class Server (ProcessClient, XMLRPCServerMixin):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)
        XMLRPCServerMixin.__init__(self)

        self.register_function(self.uptime)
        self.register_function(self.load)

    def uptime(self):
        return open('/proc/uptime').read() 

    def load(self):
        return open('/proc/loadavg').read() 

if __name__ == '__main__':
    Server(sys.argv).run()





#!/usr/bin/env python3
"""Data Transport log viewer"""

##########################################################################
#
#   ViewLog
#
#   This program displays either the server or a process group's log file
#   in a manner like "tail -F".
#
#   1.0.0   2000-??-??  Todd Valentic
#           Original implementation.
#
#   1.0.1   2002-08-26  Todd Valentic
#           Modified to be setup via configure script.
#
#   2.0.0   2005-01-07  Todd Valentic
#           Completely rewritten and no longer depends on the
#               system's tail command. Checks are made for
#               new files periodically to handle rotating log
#               files.
#
#   2.0.1   2005-04-07  Todd Valentic
#           Sort lines at start to order by time.
#
#   2.0.2   2005-11-08  Todd Valentic
#           Added key parameter to line sorting so we only sort
#               on the date.
#
#   2.0.3   2007-04-14  Todd Valentic
#           Only print log entry lines (those starting with "[")
#               at start to avoid printing part of an traceback.
#
#   2016-12-23  Todd Valentic
#               Use datatransport package format
#               Use print()
#
#   2022-10-06  Todd Valentic
#               Move into datatransport/commands
#               Convert from optparse -> argparse
#
#   2023-07-07  Todd Valentic
#               Updated for transport3 / python3
#               Add recursive and parent options
#               Handle mulitple groups/clients
#               Make sure output is always time ordered
#
#   2023-08-11  Todd Valentic
#               Show more lines at start to better catch trace backs
#
#   2026-02-26  Todd Valentic
#               Fix recursive file find. Need to use full_match instead 
#                   of match, which doesn't support the recusive wildcard
#
#               Add -V / --version option
#
##########################################################################

import argparse
import os
import signal
import sys
import time

from pathlib import Path

from datatransport import TransportConfig

VERSION = "3.0.1"

class FileTracker:
    """Track view for a given file"""

    def __init__(self, name):
        self.name = name
        self.file = None
        self.size = 0
        self.inode = 0
        self.dev = 0
        self.firsttime = True

        self.n_unchanged_stats = 0
        self.n_consecutive_size_changes = 0

        self.reopen()

    def reopen(self):
        """Reopen the file in case it closes (i.e.log rotation)"""

        try:
            # pylint: disable=consider-using-with
            stream = open(self.name, "r", encoding="utf-8")
            stats = os.stat(self.name)
        except:  # pylint: disable=bare-except
            self.file = None
            self.inode = 0
            self.dev = 0
            return

        self.n_unchanged_stats = 0
        self.n_consecutive_size_changes = 0

        if self.inode != stats.st_ino or self.dev != stats.st_dev:
            self.inode = stats.st_ino
            self.dev = stats.st_dev
            self.size = 0
            self.file = stream

            if self.firsttime:
                offset = 1000
                self.size = max(stats.st_size - offset, 0)
            else:
                self.size = 0

    def poll(self):
        """Poll files for changes"""

        if not self.file:
            self.reopen()
            return None

        try:
            stat = os.stat(self.name)
        except:  # pylint: disable=bare-except
            self.file = None
            return None

        if self.size == stat.st_size:
            self.n_unchanged_stats += 1
            self.n_consecutive_size_changes = 0
            if self.n_unchanged_stats > 5:
                self.reopen()
            return None

        self.n_unchanged_stats = 0
        self.n_consecutive_size_changes += 1

        if self.n_consecutive_size_changes > 200:
            self.reopen()
            return None

        if self.size > stat.st_size:
            self.reopen()
            return None

        self.file.seek(self.size)
        data = self.file.read()
        self.size = self.file.tell()

        if self.firsttime:
            self.firsttime = False
            if not data.startswith("\n") and "\n" in data:
                # Start at a newline
                data = data.split("\n", 1)[1]

        return data.rstrip()


def parsedate(line):
    """Get date field in line"""
    return line.split("]")[0][1:].rsplit(" ", 1)[0].strip()


class Tail:
    """Display the last lines of a groups of files"""

    def __init__(self, args):
        self.rate = 0.5
        self.args = args
        self.verbose = args.verbose
        self.specs = args.groups
        self.config = TransportConfig()
        self.trackers = {}

        self.logpath = self.config.get_path("DEFAULT", "path.logfiles")

        path = self.config.get_path("TransportServer", "log.file")
        self.serverlog = path.relative_to(self.logpath)

    def scan_groups(self, specs):
        """Scan for log files"""

        logfiles = [p.relative_to(self.logpath) for p in self.logpath.glob("**/*.log")]

        groups = []

        for spec in specs:
            path = Path(spec)
            groups.append(path)
            groups.extend(list(path.parents)[: self.args.parents])

        watch = set()

        for group in groups:
            if str(group) == ".":
                watch.add(self.serverlog)
                if self.args.recursive:
                    watch.update(logfiles)
                continue
            if self.args.recursive:
                watch.update([p for p in logfiles if p.full_match(f"{group}/**/*.log")])
            watch.update([p for p in logfiles if p.full_match(f"{group}/*.log")])
            watch.update([p for p in logfiles if p.full_match(f"{group}.log")])

        return watch

    def update_trackers(self, trackers, paths):
        """Update current log file trackers"""

        result = {}

        for path in paths:
            key = str(path)
            if key in trackers:
                tracker = trackers[key]
            else:
                tracker = FileTracker(self.logpath.joinpath(path))

            result[key] = tracker

        return result

    def run(self):
        """Print the last 10 lines of each file, sort by time"""

        if self.verbose:
            paths = self.scan_groups(self.specs)

            print("=" * 75)
            print(f"Looking in {self.specs}")
            if self.args.parents is None:
                print("Parents: all")
            else:
                print(f"Parents: {self.args.parents}")
            print(f"Total log files: {len(paths)}")
            for path in paths:
                print(path)
            print("=" * 75)

        # Tail forever

        while True:
            paths = self.scan_groups(self.specs)
            self.trackers = self.update_trackers(self.trackers, paths)

            lines = []
            for tracker in self.trackers.values():
                output = tracker.poll()
                if not output:
                    continue

                if output.startswith("["):
                    output = "\n" + output

                for entry in output.split("\n["):
                    if entry:
                        lines.append("[" + entry)

            if lines:
                lines.sort(key=parsedate)
                print(*lines, sep="\n")

            time.sleep(self.rate)


def handler(_signum, _frame):
    """Signal handler"""
    sys.exit(0)


def main():
    """Main application"""

    signal.signal(signal.SIGINT, handler)

    desc = "Data Transport Log Viewer"
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument(
        "groups", nargs="*", help="Server/Group/Client names", default="."
    )
    parser.add_argument(
        "-r", "--recursive", action="store_true", help="Show child log files"
    )
    parser.add_argument(
        "-p",
        "--parents",
        nargs="?",
        const=None,
        default=0,
        type=int,
        help="Show parent log files",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    parser.add_argument("-V", "--version", action="version", version=VERSION)

    args = parser.parse_args()

    Tail(args).run()



#!/usr/bin/env python3
"""Download messages from a newsgroup"""

#####################################################################
#
#   GetArticle
#
#   This script can be used to extract messages from a news group.
#   It can be run in a one shot or loop mode.
#
#   History:
#
#   2003-02-24  Todd Valentic
#               Initial implementation.
#
#   2004-02-17  Todd Valentic
#               Updated to new callback format.
#
#   2006-02-03  Todd Valentic
#               Removed reference for setVerbose.
#
#   2016-12-26  Todd Valentic
#               Use datatransport package
#
#   2023-08-33  Todd Valentic
#               Updated for transport3 / python3
#               Format for PEP8 compliance
#               Use argparse
#               Add signal handling
#               Add no save option
#               Add uncompress option
#               Add output directory option
#
#   2023-11-01  Todd Valentic
#               Append message ID to name when message in body
#
#   2025-04-29  Todd Valentic
#               Sometimes the attached files have the same name
#                   in each message (i.e. track.dat), resulting
#                   in the file being over written each time.
#                   To prevent this, append message ID if the 
#                   name already exists and the -m / --add-msgid
#                   option is given. 
#               Version 1.2
#
#   2025-06-14  Todd Valentic
#               Add processing script option 
#               Version 1.3
#
#####################################################################

import argparse
import bz2
import fnmatch
import gzip
import logging
import pathlib
import shutil
import signal
import subprocess
import sys
import time
import threading

from datatransport import newstool

VERSION = "1.3"


class NewsgroupPoller:
    """Monitor a single newsgroup"""

    def __init__(self, parent, newsgroup):
        self.args = parent.args
        self.log = parent.log
        self.done = parent.done
        self.newsgroup = newsgroup
        self.finished = False

        self.poller = newstool.NewsPoller()
        self.poller.set_server(self.args.server, port=self.args.port)
        self.poller.set_newsgroup(newsgroup)
        self.poller.set_stop_func(self.done.is_set)
        self.poller.set_debug(self.args.debug)
        self.poller.set_single_shot(True)
        self.poller.set_last_read_path(self.args.trackdir)
        self.poller.set_last_read_prefix(".track")
        self.poller.set_debug(True)

        self.poller.set_callback(self.process)

        if self.args.catchup:
            self.poller.mark_read(self.args.catchup)

    def uncompress(self, filename):
        """Uncompress file"""

        if not self.args.uncompress:
            return filename

        outname = filename.with_suffix("")

        if filename.suffix == ".bz2":
            handler = bz2.BZ2File
        elif filename.suffix == ".gz":
            handler = gzip.GzipFile
        else:
            return filename

        with handler(filename, "rb") as infile, outname.open("wb") as outfile:
            shutil.copyfileobj(infile, outfile)

        filename.unlink()

        return outname

    def run_script(self, script, newsgroup, timestamp, filenames):
        """Run script on filenames"""

        cmd = [script, newsgroup, f"{timestamp.isoformat()}"] + filenames
        result = subprocess.run(cmd, capture_output=True, text=True)

        if result.returncode == 0:
            self.log.info("Processing script output: %s", result.stdout)
        else:
            self.log.error("Processing script output:")
            self.log.error("  - exit code: %s", result.returncode)
            self.log.error("  - stdout: %s", result.stdout.rstrip())
            self.log.error("  - stderr: %s", result.stderr.rstrip())
            raise RuntimeError("Processing script error")

    def process(self, message):
        """Message processor"""

        newsgroup = message["Newsgroups"]
        timestamp = newstool.message_date(message)
        output = self.args.outputdir

        if self.args.addgroupdir:
            output = output.joinpath(newsgroup)

        msgnum = message["Xref"].rsplit(":", 1)[-1]
        filenames = newstool.save_files(
            message, write=self.args.save, path=output
        )
        self.log.info("Received files from %s", message["Newsgroups"])

        saved_filenames = []

        for filename in filenames:

            if self.args.add_msgnum:
                # note - there might be multiple suffixess
                stem, suffix = filename.name.split(".", 1)
                newname = f"{stem}-{msgnum}"
                if suffix:
                    suffix = "." + suffix 
                newname = filename.parent / f"{stem}-{msgnum}{suffix}"
                filename.rename(newname)
                filename = newname

            filename = self.uncompress(filename)
            self.log.info("  - %s", filename.name)

            saved_filenames.append(filename)

        if self.args.script and saved_filenames:
            self.log.debug("  - running %s", self.args.script)
            self.run_script(self.args.script, newsgroup, timestamp, saved_filenames)

        if self.args.oneshot:
            self.finished = True

    def run(self):
        """Poll newsgroup for new messages"""

        self.log.debug("Checking %s", self.newsgroup)
        self.poller.poll()
        return not self.finished


class Downloader:
    """Download articles from multiple newsgroups"""

    def __init__(self):
        self.setup_signals()

        self.done = threading.Event()
        self.args = self.parse_command_line()
        self.log = self.setup_log(self.args)

        self.pollers = self.create_pollers(self.args)

        self.show_config()

    def setup_signals(self):
        """Setup signal handlers"""

        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGHUP, self.signal_handler)

    def signal_handler(self, _signum, _frame):
        """Signal handler"""

        self.done.set()

    def show_config(self):
        """Display banner and configuration options"""

        self.log.info("=" * 75)
        self.log.info("Data Transport News Article Downloader %s", VERSION)
        self.log.info("=" * 75)

        self.log.info("Press Ctrl-C to exit")
        self.log.info("")

        self.log.info("News server: %s:%s", self.args.server, self.args.port)
        self.log.info("One shot: %s", self.args.oneshot)
        self.log.info("Catchup: %s", self.args.catchup)
        self.log.info("Save files: %s", self.args.save)
        self.log.info("Output path: %s", self.args.outputdir)
        self.log.info("Run script: %s", self.args.script)
        self.log.info("")

        self.log.info("Watching newsgroups:")

        for poller in self.pollers.values():
            self.log.info("  %s", poller.newsgroup)

        self.log.info("-" * 75)

    def parse_command_line(self):
        """Parse the command line"""

        desc = "Download messages from a newsgroup" ""
        parser = argparse.ArgumentParser(description=desc)

        parser.add_argument("-V", "--version", action="version", version=VERSION)

        parser.add_argument(
            "-q", "--quiet", action="store_true", help="Only show errors"
        )

        parser.add_argument("-d", "--debug", action="store_true", help="Debug output")

        parser.add_argument(
            "-c",
            "--catchup",
            type=int,
            nargs="?",
            default=0,
            help="Mark all messages as read, -N mark up to last N messages",
        )

        parser.add_argument(
            "-e",
            "--oneshot",
            action="store_true",
            help="Exit after one message is received from each newsgroup",
        )

        parser.add_argument(
            "-m",
            "--add_msgnum",
            action="store_true",
            dest="add_msgnum",
            help="Append message number to output filenames",
        )

        parser.add_argument(
            "-n",
            "--nosave",
            action="store_false",
            dest="save",
            help="Don't save files, just report",
        )

        parser.add_argument(
            "-u",
            "--uncompress",
            action="store_true",
            help="Uncompress gz or bz2 files",
        )

        parser.add_argument(
            "-o",
            "--outputdir",
            default=".",
            type=pathlib.Path,
            help="Store files in this directory",
        )

        parser.add_argument(
            "-g",
            "--addgroupdir",
            action="store_true",
            help="Add newsgroup name to output directory",
        )

        parser.add_argument(
            "-t",
            "--trackdir",
            default=".",
            type=pathlib.Path,
            help="Store tracking files in this directory",
        )

        parser.add_argument(
            "-s",
            "--server",
            default="localhost",
            help="News server (default: localhost)",
        )

        parser.add_argument(
            "-p",
            "--port",
            type=int,
            default=119,
            help="News server port (default: 119)",
        )

        parser.add_argument(
            "-r",
            "--run",
            dest="script",
            help="Run script on output files",
        )

        parser.add_argument("newsgroups", nargs="+", metavar="newsgroup")

        args = parser.parse_args()

        if args.catchup is None:
            args.catchup = 1

        return args

    def setup_log(self, args):
        """Setup the log output"""

        fmt = "[%(levelname)s] %(message)s"

        if args.debug:
            logging.basicConfig(format=fmt, level=logging.DEBUG)
        elif args.quiet:
            logging.basicConfig(format=fmt, level=logging.ERROR)
        else:
            logging.basicConfig(format=fmt, level=logging.INFO)

        return logging

    def create_pollers(self, args):
        """Create a poller for each newsgroup"""

        server = newstool.NewsTool()
        server.set_server(args.server, port=args.port)
        available_newsgroups = server.list_newsgroups()

        pollers = {}

        for pattern in args.newsgroups:
            for newsgroup in available_newsgroups:
                if fnmatch.fnmatch(newsgroup, pattern):
                    if newsgroup not in pollers:
                        pollers[newsgroup] = NewsgroupPoller(self, newsgroup)

        return pollers

    def polling_cycle(self, pollers):
        """Run a polling cycle for all of the monitored newsgroups"""

        self.log.debug("Cycle start")

        next_pollers = []

        for poller in pollers:
            if poller.run():
                next_pollers.append(poller)

        if self.args.oneshot and pollers:
            self.log.debug("Waiting for messages from")
            for poller in next_pollers:
                self.log.debug("  %s", poller.newsgroup)

        self.log.debug("Cycle finished")

        return next_pollers

    def run(self):

        pollers = self.pollers.values()
        exit_code = 0

        while not self.done.is_set() and pollers:
            try:
                pollers = self.polling_cycle(pollers)
            except Exception:
                exit_code = 1
                break

            time.sleep(1)

        return exit_code 

    

def main():
    """Script entry point"""

    return_code = Downloader().run()
    sys.exit(return_code)



#!/usr/bin/env python
"""Remove files"""

########################################################################
#
#   remove_file
#
#   This is a small routine to remove files. It first tests to see if
#   the file exists and then tries to remove it. This approximates the
#   'rm -f' command. The input parameter can be a single filename or a
#   list of filenames. Each name can be a string, bytes or a pathlib
#   object.
#
#   2001-08-27  Todd Valentic
#               Initial implementation.
#
#   2001-10-11  Todd Valentic
#               Modified to handle a list of filenames.
#
#   2022-10-10  Todd Valentic
#               Python3 port
#               Support pathlib
#               removeFile -> remove_file
#
#   SPDX-License-Identifier: GPL-3.0-or-later
#   Copyright (C) 1999-2022 Todd Valentic
#
########################################################################

import pathlib

from collections.abc import Iterable
from typing import Union

def remove_file(filenames: Union[str, bytes, Iterable]) -> None:
    """
    Remove a filename or list of filenames. Ignore errors.
    """

    if isinstance(filenames, (str, bytes)):
        filenames = [filenames]
    elif not isinstance(filenames, Iterable):
        filenames = [filenames]

    for filename in filenames:
        path = pathlib.Path(str(filename))
        path.unlink(missing_ok=True)



###################################################################
#
#   Stand alone access to transport directory services
#
#   2005-11-11  Todd Valentic
#               Initial implementation.
#
#   2023-10-06  Todd Valentic
#               Updated for python3 / transport3
#
###################################################################

import xmlrpc.client

class Directory:
    """Connect to xmlrpc directory service"""

    def __init__(self,host='localhost', port=8411, service=None):

        kwargs = {"allow_none": True, "use_builtin_types": True}

        url = 'http://%s:%d' % (host,port)
        self.directory = xmlrpc.client.ServerProxy(url, **kwargs)

        if service:
            self.directory = self.connect(service)

    def list(self):
        return self.directory.list()

    def connect(self, service, **kwargs):
        """Connect to a service via the directory"""

        args = {"allow_none": True, "use_builtin_types": True}

        args.update(kwargs)

        url = self.directory.get(service, "url")
        server  = xmlrpc.client.ServerProxy(url, **args)

        return server

        




#!/usr/bin/env python

#####################################################
#
#   SBC Plotter
#
#   Example of writing a data format helper
#
#   2010-08-21  Todd Valentic
#               Initial implementation
#
#####################################################

from Transport.Components   import PlotTool

import sys
import ConfigParser

class PlotSBC(PlotTool):

    def parse(self,message):

        # body is one line long, format:
        #
        #   [fan] 0

        body = message.get_payload()

        self.data = ConfigParser.ConfigParser()
        self.data.add_section('State')

        body = body.split('\n')
        unit,state = body[0].split()

        # [fan] -> fan

        unit = unit[1:-1]

        self.data.set('State',unit,state)

        return True

if __name__ == '__main__':
    PlotSBC(sys.argv).run()



#!/usr/bin/env python

from Transport	import ProcessClient

import sys

class Demo (ProcessClient):

	def __init__(self,args):
		ProcessClient.__init__(self,args)

		self.log.info('space.none:    %d' % self.getBytes('space.none'))
		self.log.info('space.bytes:   %d' % self.getBytes('space.bytes'))
		self.log.info('space.kbytes:  %d' % self.getBytes('space.kbytes'))
		self.log.info('space.mbytes:  %d' % self.getBytes('space.mbytes'))
		self.log.info('space.gbytes:  %d' % self.getBytes('space.gbytes'))
		
		self.log.info('nospace.none:    %d' % self.getBytes('nospace.none'))
		self.log.info('nospace.bytes:   %d' % self.getBytes('nospace.bytes'))
		self.log.info('nospace.kbytes:  %d' % self.getBytes('nospace.kbytes'))
		self.log.info('nospace.mbytes:  %d' % self.getBytes('nospace.mbytes'))
		self.log.info('nospace.gbytes:  %d' % self.getBytes('nospace.gbytes'))

		self.log.info('bad.none:    %d' % self.getBytes('bad.none'))
		self.log.info('bad.bytes:   %d' % self.getBytes('bad.bytes'))
		self.log.info('bad.kbytes:  %d' % self.getBytes('bad.kbytes'))
		self.log.info('bad.mbytes:  %d' % self.getBytes('bad.mbytes'))

	def run(self):

		while self.wait(60):
			pass

if __name__ == '__main__':
	Demo(sys.argv).run()



#!/usr/bin/env python
"""Transport Config"""

#############################################################################
#
#   Transport Config
#
#   This object provides a simple interface to load the global transport
#   configuration file.
#
#   Revision history:
#
#   1.0.0   2000-06-19  TAV
#           Initial implementation
#
#   1.0.1   2001-09-03  TAV
#           Added the getUserInfo() method to return uid/gid.
#
#   1.0.2   2002-08-27  Todd Valentic
#           Setup by configure script.
#           Read hostname and user/group info from config file.
#
#   1.0.3   2004-08-08  Todd Valentic
#           Added umask.
#           Use DefaultConfigParser
#
#   1.0.4   2005-01-06  Todd Valentic
#           Changed DefaultConfigParser to be a local import
#
#   1.0.5   2006-06-03  Todd Valentic
#           Moved default location from server -> etc
#
#   1.0.6   2009-09-07  Todd Valentic
#           Read other .conf files in main config directory
#
#   1.0.7   2019-04-10  Todd Valentic
#           Lookup uid/gid at start instead of from config file
#
#   2016-12-23  Todd Valentic
#               Use DATA_TRANSPORT_PATH environment variable
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#
#   2023-06-27  Todd Valentic
#               Use SapphireConfigParser
#               Remove unused abort() method
#               Remove uid, gid, username, groupname, hostname, umask
#
#   2023-07-09  Todd Valentic
#               Add find_config_files() method
#
#   2025-04-15  Todd Valentic
#               Ensure main transportd.conf file is present
#
#############################################################################

import errno
import os
from pathlib import Path 

import sapphire_config as sapphire


class TransportConfig(sapphire.Parser):
    """Transport configuration"""

    def __init__(self, defaults=None):
        prefix = os.environ.get("DATA_TRANSPORT_PATH", "/opt/transport")
        confdir = Path(prefix, "etc")
        mainconf = confdir.joinpath("transportd.conf")

        if not mainconf.exists():
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                str(mainconf)
            )

        filenames = sorted(confdir.glob("*.conf"))

        # make sure transportd.conf is read first

        filenames.remove(mainconf)
        filenames.insert(0, mainconf)

        super().__init__(defaults=defaults)

        self.read(filenames)

    def find_config_files(self, groupname, basepath, hostname):

        def glob_conf(curpath, ext):
            mainconf = curpath.joinpath(f"{curpath.name}.{ext}")
            paths = sorted(list(curpath.glob(f"*.{ext}")))

            if mainconf in paths:
                paths.remove(mainconf)
                paths.insert(0, mainconf)

            return paths

        configpaths = []

        for path in reversed(Path(groupname,'x').parents):
            curpath = basepath.joinpath(path)

            configpaths.extend(glob_conf(curpath, 'conf'))
            configpaths.extend(glob_conf(curpath, f'conf-{hostname}'))

        return configpaths




from .metadata import __version__

# Order is important to avoid circular import

from .root import Root
from .transportconfig import TransportConfig
from .configcomponent import ConfigComponent
from .processclient import ProcessClient
from .accessmixin import AccessMixin
from .directory import Directory
from .processgroup import ProcessGroup
from .transportserver import TransportServer
from .transportmanager import TransportManager
from .xmlrpcserver import XMLRPCServer
from .newsposter import NewsPoster 
from .newspoller import NewsPoller



#!/usr/bin/env python2

############################################################################
#
#   Timer   
#
#   Test wait() timing. 
#
#   History:
#   
#   1.0.0   2005-09-19  Todd Valentic   
#           Initial implementation
#
############################################################################

from    Transport   import ProcessClient

import  sys

class Timer(ProcessClient):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)

    def run(self):

        rate   = self.getDeltaTime ('poll.rate',60)
        sync   = self.getboolean   ('poll.sync',True)
        offset = self.getDeltaTime ('poll.offset',0)

        while self.running:
            self.log.info('Waiting')
            self.wait(rate,offset=offset,sync=sync)
            self.log.info('  mark')

if __name__ == '__main__':
    Timer(sys.argv).run()




#!/usr/bin/env python

##########################################################
#
#   Example 3 - poll a news group
#
#   Print any messages from the news group to the log
#
#   2010-08-16  Todd Valentic
#               Initial implementation
#
##########################################################

from Transport import ProcessClient
from Transport import NewsPollMixin

import sys

class PollExample(ProcessClient,NewsPollMixin):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)
        NewsPollMixin.__init__(self,callback=self.process)

    def process(self,message):

        body = message.get_payload()

        self.log.info('Received: %s' % body)

if __name__ == '__main__':
    PollExample(sys.argv).run()




#!/usr/bin/env python3
"""Newsgroup Monitor"""

#####################################################################
#
#   Newsgroup Monitor
#
#   Monitors the news groups for messages
#
#   2009-05-04  Todd Valentic
#               Initial implementation. Based on InstrumentStatus
#
#   2010-03-20  Todd Valentic
#               Added feedgroups
#
#   2010-03-23  Todd Valentic
#               Save history at start
#
#   2011-03-11  Todd Valentic
#               Add warning state
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#               Python3 updates:
#                   StringIO -> io.StringIO
#               NewsPoller, NewsPoster
#
#   2023-07-29  Todd Valentic
#               Updated for transport3 / python3
#
#####################################################################

import io
import sys

from datatransport import ProcessClient
from datatransport import NewsPoster
from datatransport import NewsPoller
from datatransport import ConfigComponent
from datatransport.apps import EventMonitor

class Feed(EventMonitor.Member):
    """Feed event monitor"""

    def __init__(self, *pos, **kw):
        EventMonitor.Member.__init__(self, "feed", *pos, **kw)

        self.news_poster = NewsPoster(self, quiet=True)
        self.news_poller = NewsPoller(self)
        self.main = news_poller.main

        self.add_states("online", "offline", "warning", "error")

        self.feedgroup = kw["localvars"]["feedgroup"]

        if self.feedgroup:
            desc = "%s:%s" % (self.feedgroup, self.name)
        else:
            desc = self.name

        self.log.info("    %s:%s" % (self.feedgroup, self.name))

        self.timeout = self.get_timedelta("timeout", "1:00:00")
        self.report_change = self.get_boolean("report.change", True)
        self.desc = self.get("desc", desc)
        self.alwayson = self.get_boolean("alwayson", False)

    def check(self):

        groups = {}

        for poller in self.news_pollers:
            groups.update(poller.list_articles(self.timeout))

        count = 0
        feeds = 0

        for group, articles in groups.items():
            if articles:
                feeds += 1
                count += len(articles)

        if feeds:
            if len(groups) == feeds:
                self.update("online")
            else:
                self.update("warning")
        else:
            if self.alwayson:
                self.update("error")
            else:
                self.update("offline")

        msg = "  - %s: %s %s (%s msgs in %s feeds over the last %s)" % (
            self.name,
            self.cur_state,
            self.current.elapsed,
            count,
            feeds,
            self.timeout,
        )

        if self.changed:
            msg += " <- Changed from %s" % self.prev_state

        self.log.info(msg)

        if self.report_change and self.changed:
            self.report()

    def report(self):

        subject, message = self.report_message()

        try:
            self.news_poster.set_subject(subject)
            self.news_poster.post_text(message)
        except:
            self.log.exception("Error posting change message for %s" % self.name)

    def report_message(self):

        if self.cur_state == "online":
            action = "started"
        else:
            action = "stopped"

        subject = "%s has %s" % (self.desc, action)

        fmt = "%a, %d %b %Y %H:%M:%S %Z"

        message = []
        message.append("")
        message.append("  The following instrument has %s sending data:" % action)
        message.append("")
        message.append("    Instrument: %s" % self.desc)
        message.append("    Date/Time : %s" % self.current.starttime.strftime(fmt))
        message.append("    Currently : %s" % self.cur_state.capitalize())
        message.append("")
        message.append("  Previous state: %s" % self.prev_state.capitalize())
        message.append("    Start time: %s" % self.previous.starttime.strftime(fmt))
        message.append("    Stop time : %s" % self.previous.lasttime.strftime(fmt))
        message.append("    Elapsed   : %s" % self.previous.elapsed)
        message.append("")

        message = "\n".join(message)

        return subject, message


class FeedGroup(ConfigComponent):
    """Feed group component"""

    def __init__(self, name, config, parent):
        ConfigComponent.__init__(self, "feedgroup", name, config, parent)

        self.log.info("  Feed group: %s", name)

        localvars = {"feedgroup": name, "section": "%(feedgroup)s:%(feed)s"}

        # TBD - Do we really need to forward there vars?

        for key, option in self.optionsdict().items():
            value = self.get(option, raw=True)
            localvars["feedgroup." + key] = value

        self.feeds = self.config.get_components("feeds", factory=Feed, vars=localvars)

    def check(self):

        for feed in self.feeds.values():
            try:
                feed.check()
            except:
                self.log.exception("Problem checking %s", feed.name)

class NewsgroupMonitor(ProcessClient):
    """Process client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.history_poster = NewsPoster(self, prefix="history", quiet=True)
        self.status_poster = NewsPoster(self, prefix="status", quiet=True)
        self.update_poster = NewsPoster(self, prefix="update", quiet=True)

        self.log.info("Loading feed groups")

        self.feedgroups = self.config.get_components("feedgroups", factory=FeedGroup)

        if not self.feedgroups:
            self.feedgroups = [FeedGroup("", self)]

        self.feeds = {}

        for feedgroup in self.feedgroups.values():
            for feed in feedgroup.feeds.values():
                key = "%s:%s" % (feedgroup.name, feed.name)
                self.feeds[key] = feed

        self.tracker = EventMonitor.EventMonitor(self.feeds)
        self.tracker.load("tracker")

        self.first_time = True

    def main(self):

        rate = self.get_timedelta("poll.rate", "1:00")
        sync = self.get_boolean("poll.sync")
        offset = self.get_timedelta("poll.offset")

        while self.wait(rate, sync, offset):

            for feedgroup in self.feedgroups:
                self.log.info("Checking feeds: %s" % feedgroup.name)
                feedgroup.check()

            try:
                self.save_state()
            except:
                self.log.exception("Problem saving tracker")

    def save_state(self):

        config = self.tracker.save("tracker")
        strbuff = io.StringIO()
        config.write(strbuff)

        self.update_poster.post_text(strbuff.getvalue())

        if self.first_time:
            self.history_poster.post_text(strbuff.getvalue())
            self.first_time = False

        if self.tracker.changed():
            self.history_poster.post_text(strbuff.getvalue())

            subject, message = self.status_message()
            self.status_poster.set_subject(subject)
            self.status_poster.post_text(message)

    def status_message(self):

        message = []
        message.append("")
        message.append("  A change in instrument status has been detected:")
        message.append("")

        subject = None

        descwidth = max([len(feed.desc) for feed in self.feeds.values()])
        fmt = "%a, %d %b %Y %H:%M:%S %Z"

        for feed in self.feeds.values():

            status = feed.cur_state.capitalize()

            if not feed.changed:
                changed = ""
            else:
                changed = "<-- %s" % feed.current.starttime.strftime(fmt)
                if not subject:
                    subject = "%s %s" % (feed.desc, feed.cur_state)
                else:
                    subject = "Multiple instruments have changed status"

            message.append(
                "  %s: %-10s %s" % (feed.desc.rjust(descwidth), feed.cur_state, changed)
            )

        message.append("")
        message = "\n".join(message)

        return subject, message


def main():
    NewsgroupMonitor(sys.argv).run()



#!/usr/bin/env/python3
"""File Post Component"""

##########################################################################
#
#   FilePost
#
#   This transport component watches for specified files to appear and
#   posts them into a newsgroup. If the newsgroup does not exist, it
#   will be created. Multiple files can be listed and they will be posted
#   as a group of attachments in a single message. This is a useful
#   feature if you have data files that belong together and should stay
#   bundled in the transport network. Once the files are posted, they are
#   removed. An option exists to leave the files as well. For the files
#   to be removed, they need to have write permission for the transport user.
#
#   The component is based on the original FileWatch componment. It
#   expands on it by allowing you to specify multiple groups of files
#   to post, each with its own destination news group. This approach
#   scales much better when there are a number of groupings.
#
#   It uses the standard transport ConfigComponent to specify the
#   groups. Each grouping can list:
#
#       post.newsgroup (and other related options for NewsPoster)
#       remove_files - remove files after posting (default True)
#       group_files - post all files to one message (default False)
#       files - list of file names (including path) to look for. If
#               multiple names are listed, they should be separated
#               by spaces. The name can include normal shell wildcards.
#
#   2010-08-16  Todd Valentic
#               Initial implementation
#
#   2017-01-11  Todd Valentic
#               Handle empty filegroup list
#
#   2022-10-12  Todd Valentic
#               Python3 port
#                   Transport -> datatransport
#                   removeFile -> remove_file
#                   NewsPoster
#
#   2023-07-27  Todd Valentic
#               Updated for transport3 / python3
#
##########################################################################

import sys

from datatransport import ProcessClient
from datatransport import NewsPoster
from datatransport import ConfigComponent
from datatransport.utilities import remove_file


class Watcher(ConfigComponent):
    """Watch group component"""

    def __init__(self, *p, **kw):
        ConfigComponent.__init__(self, "watch", *p, **kw)

        self.news_poster = NewsPoster(self)

        self.watchpath = self.config.get_path("path", ".")
        self.filespecs = self.config.get_list("files")
        self.remove_files = self.config.get_boolean("removefiles", True)
        self.group_files = self.config.get_boolean("groupfiles", False)

        if not self.filespecs:
            self.abort("No watch files listed in the config file")

        self.log.info("Posting to %s", self.config.get("post.newsgroup"))
        self.log.info("Watching path: %s", self.watchpath)
        self.log.info("Watching for: %s", self.filespecs)

    def find_files(self):
        """Find matching files to post"""

        filenames = []

        for filespec in self.filespecs:
            filenames.extend(self.watchpath.glob(filespec))

        filenames = [f for f in filenames if f.is_file()]

        return sorted(filenames)

    def process(self):
        """Check for new files"""

        filenames = self.find_files()

        if not filenames:
            self.log.debug("No files present")
            return

        if self.group_files:
            filenames = [filenames]

        for filegroup in filenames:
            starttime = self.now()

            try:
                self.news_poster.post(filegroup)
            except Exception as e:  # pylint: disable=broad-exception-caught
                self.log.exception("Problem posting files: %s", e)
                return

            elapsed = (self.now() - starttime).total_seconds()
            
            if isinstance(filegroup, list):
                names = [str(n) for n in filegroup]
            else:
                names = filegroup

            self.log.info("Files posted (%0.2fs): %s", elapsed, names)

            if self.remove_files:
                try:
                    remove_file(filegroup)
                except OSError as e:
                    self.log.exception("Problem deleting file %s: %s", filegroup, e)


class FilePost(ProcessClient):
    """Process Client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.rate = self.config.get_rate("pollrate")
        self.watches = self.config.get_components("watches", factory=Watcher)

    def process(self):
        """Process each file watch group"""

        for watch in self.watches.values():
            watch.process()

            if self.is_stopped():
                break

    def main(self):
        """Main application"""

        while self.wait(self.rate):
            self.process()


def main():
    """Script entry point"""
    FilePost(sys.argv).run()



#!/usr/bin/env python3
"""Generate source messages"""

import sys

from datatransport import ProcessClient
from datatransport import NewsPoster


class Source(ProcessClient):
    """Source Process Client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.rate = self.config.get_rate("post.rate", 60)
        self.poster = NewsPoster(self)

        self.log.info("rate: %s", self.rate)

    def main(self):
        """Main application"""

        while self.wait(self.rate):
            msg = f"[self.name] {self.now()}"
            self.poster.post_text(msg)
            self.log.info("Posted message")


if __name__ == "__main__":
    Source(sys.argv).run()



#!/usr/bin/env python

###########################################################
#
#   XMLRPC Client Program
#
#   How to user an XMLRPC service
#
#   2010-08-19  Todd Valentic
#               Initial implementation
#
###########################################################

from Transport  import ProcessClient
from Transport  import DirectoryMixin

import sys

class Client (ProcessClient, DirectoryMixin):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)
        DirectoryMixin.__init__(self)

        self.systemstatus = self.connect('systemstatus')

    def run(self):

        while self.wait(5):
            load = self.systemstatus.load()
            uptime = self.systemstatus.uptime()

            self.log.info('Load: %s' % load)
            self.log.info('Uptime: %s' % uptime)


if __name__ == '__main__':
    Client(sys.argv).run()



#!/usr/bin/env python2


from Transport import ProcessClient
from Transport import NewsPollMixin

class CheckDate(ProcessClient,NewsPollMixin):

	def __init__(self,argv):
		ProcessClient.__init__(self,argv)
		NewsPollMixin.__init__(self,processFiles=self.process)

	def process(self,

if __name__ == '__main__':
	CheckDate(sys.argv).run()



#!/usr/bin/env python

############################################################################
#
#   Config  
#
#   This script tests config file access. 
#
#   History:
#   
#   1.0.0   2005-11-12  Todd Valentic   
#           Initial implementation
#
############################################################################

from    Transport   import ProcessClient

import  sys

class Config (ProcessClient):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)

    def run(self):

        self.log.info('missing: %s' % self.get('missing','OK: The default value'))
        self.log.info('default: %s' % self.get('value.default'))
        self.log.info('local:   %s' % self.get('value.local'))
        self.log.info('var:     %s' % self.get('value.label',vars={'name':'world'}))

        self.log.info('Finished')

if __name__ == '__main__':
    Config(sys.argv).run()




#!/usr/bin/env python2

############################################################################
#
#	CreateData	
#
#	This script creates test data files. 
#
#	History:
#	
#	1.0.0 	2002-12-12	Todd Valentic	
#			Initial implementation
#
############################################################################

from 	Transport	import ProcessClient

import	os
import	sys
import	random

class CreateData(ProcessClient):

	def __init__(self,argv):

		ProcessClient.__init__(self,argv)

		self.rate 	= self.getDeltaTime('rate',60)
		self.path	= self.get('output.path','')
		self.names	= self.get('output.names','').split()

		if self.names==[]:
			self.abort('You need to specify at least one output name (output.names)')

	def run(self):

		while self.running:

			for name in self.names:

				filename = os.path.join(self.path,name)

				path = os.path.dirname(filename)

				if path and not os.path.exists(path):
					try:
						os.makedirs(path)
					except:
						self.log.exception('Problem creating output path')
						self.abort('Exiting')
				
				try:
					numbytes = int(random.random()*10000)
					open(filename,'w').write('.'*numbytes)
					self.log.info('Created file: %s (%d bytes)'%(filename,numbytes))
				except:
					self.log.exception('Problem creating file')
	
				self.wait(self.rate)


if __name__ == '__main__':
	CreateData(sys.argv).run()




#!/usr/bin/env python3
"""Create a newsgroup"""

#####################################################################
#
#   CreateNewsgroup
#
#   This script creates a newsgroup on the news server.
#
#   2002-03-04	Todd Valentic
#               Initial implementation
#
#   2016-12-25  Todd Valentic
#               Use datatransport package
#
#   2023-08-23  Todd Valentic
#               Updated for transport3 / python3
#               Add command line options
#
#   2023-11-09  Todd Valentic
#               Add filename option
#               Version 1.1
#
#####################################################################

import argparse
import sys
import time

from datatransport import newstool

VERSION = "1.1"


def main():
    """Script entry point"""

    desc = "Remove a newsgroup from the news server"
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument("-V", "--version", action="version", version=VERSION)

    parser.add_argument(
        "-s", "--server", default="localhost", help="News server (default: localhost)"
    )
    parser.add_argument(
        "-p", "--port", default=119, type=int, help="News server port (default: 119)"
    )
    parser.add_argument(
        "-f", "--filename", help="Read newsgroup names from file"
    )

    parser.add_argument("newsgroups", nargs="*", metavar="newsgroup")

    args = parser.parse_args()

    newsgroups = []
    newsgroups.extend(args.newsgroups)

    if args.filename:
        with open(args.filename, encoding="utf8") as f: 
            for line in f:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                newsgroups.append(line)

    if not newsgroups: 
        print("No newsgroups listed")
        sys.exit(0)

    server = newstool.NewsControl()
    server.set_server(args.server, port=args.port)

    for newsgroup in newsgroups:
        if server.has_newsgroup(newsgroup):
            print(f"The newsgroup already exists: {newsgroup}")
            continue

        print(f"Creating the newsgroup: {newsgroup}")

        server.newgroup(newsgroup)

        print("Waiting for group to show up:  ", end="", flush=True)

        deadline = time.time() + 120

        wheel = ["-", "/", "|", "\\"]
        curpos = 0

        while not server.has_newsgroup(newsgroup):
            print(f"\b{wheel[curpos]}", end="", flush=True)
            curpos = (curpos + 1) % len(wheel)
            time.sleep(1)

            if time.time() > deadline:
                print("")
                print("Timeout. Failed to create newsgroup")
                sys.exit(1)

        print("\bDone")

    sys.exit(0)



#!/usr/bin/env python3

##########################################################################
#
#   Polling example
#
#   Periodically poll for messages from a news group 
#
#   2022-10-10  Todd Valentic
#               Initial implementation
#
##########################################################################

from datatransport import ProcessClient
from datatransport import NewsPollMixin 
from datatransport import newstool
from datatransport.utilities import remove_file

import sys

class Client (ProcessClient, NewsPollMixin):

    def __init__(self, args):
        ProcessClient.__init__(self, args)
        NewsPollMixin.__init__(self, callback=self.process)

        self.log.info(f'Poll for messages')

    def process(self, message):

        filenames = newstool.save_files(message)
        self.log.info(f'Received files: {[str(f) for f in filenames]}')

        remove_file(filenames)

if __name__ == '__main__':
    Client(sys.argv).run()




#!/usr/bin/env python3
"""ArchiveGroups Component"""

# pylint: disable=bare-except

##############################################################################
#
#  ArchiveGroups
#
#  Archives messages held in multiple newsgroups.
#
#  History:
#
#   1.0.0   2000-11-03  TAV
#           Initial implementation
#
#   1.0.1   2000-11-16  TAV
#           Use fnmatch instead of regular-expression matching.
#
#   1.0.2   2000-08-07  TAV
#           Code cleanup.
#           Added better error checking and message printing.
#           Try to use builtin functions when possible (limit os.system())
#
#   1.0.3   2001-09-05  TAV
#           Made groupExp (config file's pollgroups) a list of group
#               expressions. The old config files are still compatible,
#               but you can now list multiple patterns to watch for.
#
#           Added more comple <group> pattern replacement. You can now
#               give a list slice expression to pick out parts of the
#               group name. For example, if the group is transport.test.data
#               then the patter <group[1]> would return "test". A typical
#               usage is to extract the last part of the group name:
#               <group[-1]> returns "data". This feature is implemented by
#               splitting the group name up at the "." boundries and
#               evaluting the following list with the pattern.
#
#           Modified the <group> pattern to have <procgroup..> and
#               <newsgroup..>.
#
#           Switch to mx.DataTime to avoid localtime issues.
#           Added maxMessage check to limit number of files that are
#               concatenated (settng to 1 is equivalent to no concatenation
#               and a max of 0 means no limit).
#
#   1.0.4   2001-09-09  TAV
#           Created PatternTemplate class to extract pattern replacement
#               logic.
#
#           Added ability to enable report posting based on different types
#               of events (max messages, time out, time gap).
#
#   1.0.5   2001-09-13  TAV
#           Added matching rule feature.
#
#   2.0.0   2001-10-03  TAV
#           Complete rewrite of the script. It now handles the duties of a
#               number of other scripts dealing with the storing of messages
#               from the newsserver. It handles concatenated record storage
#               as well as file history and expiration. Notification messages
#               can be send on different events to a newsgroup to inform
#               others when data is ready or a newsgroup has stopped sending
#               data.
#
#   2.2.0   2001-10-17  TAV
#           Another major rewrite. I think I've got it down now. Correctly
#               handles archiving multi-part messages. Simplified things
#               considerably.
#
#   2.2.1   2001-11-06  TAV
#           Fixed a syntax error calling touch_message().
#
#   2.2.2   2001-11-08  TAV
#           Fixed an error in roll_files(). The last file was being removed
#               even if history was set to be 0 (which made the behavior
#               the same as history=1).
#
#   2.2.3   2001-11-09  TAV
#           Renamed NewsTool -> NewsTool
#
#   2.2.4   2001-11-13  TAV
#           Instrumented save_file() routine to better tell what happens it
#               fails.
#
#   2.2.5   2001-11-28  TAV
#           Modified the article processing routine to correctly handle
#               messages that are not in a multipart MIME format (ie plain
#               text messages). These are stored into a file named "noname"
#               and then processed as usual.
#
#   2.2.6   2001-11-29  TAV
#           Improved date handling. Originally, it was assumed that messages
#               had the data encoded in the Subject line (in the manner the
#               NewsTool.py employs). However, this is not always the case.
#               We used to just bail if we couldn't get the date. Now, a
#               set of dates are checked:
#                   - Subject line
#                   - NNTP-Posting-Date
#                   - Date
#                   - If those fail, use current time()
#
#   2.2.7   2001-11-30  TAV
#           Made the file name for plain text (see comments in 2.2.5) to be
#               a config parameter and set the default to noname.txt.
#
#   2.2.8   2001-12-04  TAV
#           Changed report.enable* defaults to off.
#           Fixed problem that kept the message number from being incremented
#               when no files were present in the message.
#
#   2.2.9   2001-12-10  TAV
#           Added ability to uncompress files during save if they were
#               compressed witg gzip and have a ".gz" extension. This
#               feature can be enabled/disabled with the output.uncompress
#               config option and defaults to disabled. In the future I'd
#               like to expand this to include .zip and tarballs as well.
#           Fixed date problem in filenames.
#
#   2.2.10  2001-12-26  TAV
#           Changed tav -> sri
#
#   2.2.11  2002-12-14  TAV
#           Added summary report generation. This addition requires new
#               checkpoint databases because the format changed.
#
#   2.2.12  2001-12-15  TAV
#           Fixed bug in default rules specification. The output name
#               should be the filename: (*,*,*) -> (*,*,<filename>)
#
#   2.2.13  2002-01-22  TAV
#           No longer try to set the group id when making files or dirs.
#               This is better handled using the user private group scheme
#               and saves headaches here. No longer need makePath().
#
#   2.2.14  2002-01-23  TAV
#           Added ability to disable the summary file tracking and limit
#               the maximum number of files being kept. If the summaries
#               are not being used, the file log was never cleared and
#               the database quickly grew to an enormous size, leading to
#               the crash of ArchiveGroups.
#
#           Started to add traceback printing to log on errors.
#           Signal handling now done in ProcessClient.
#
#   2.2.15  2002-02-19  Todd Valentic
#           Fixed bugs in calls to checkpoint.touch_message()
#
#   2.2.16  2002-08-028 Todd Valentic
#           sri.transport -> Transport
#           Fixed problem with timegap being reported after timeout.
#           Made separate report and summary news posters.
#           Track files in report mode if timeout or timegap is enabled.
#
#   2.2.17  2002-10-08  Todd Valentic
#           Set umask to 0 before creating path. The mode wasn't being
#               set properly otherwise.
#
#   2.2.18  2002-10-13  Todd Valentic
#           Added save_callback feature.
#
#   2.2.19  2003-03-20  Todd Valentic
#           Added ability to add an offset to message dates. Use this when
#               you need to "adjust" the date. For example, the timestamps
#               might be in UTC, but you want the files stamed in local time.
#               new time = message time + offset
#
#   2.2.20  2003-04-10  Todd Valentic
#           NewsToolBase -> NewsTool
#
#   2.2.21  2003-06-06  Todd Valentic
#           Added try..except around group check. We were dying if the
#               news server refused new connections (happens during high
#               load situations like updatedb).
#
#   2.2.22  2003-06-11  Todd Valentic
#           Added callback feature. This lets you modify the file and/or
#               filename prior to archive. This also involved changes
#               the order in which files are saved: all of the files
#               in a messages are now first uncompressed, then processed
#               then saved.
#           The check for compression was looking for gzip - right now
#               only bzip is supported.
#           Use the new NewsTool.SaveFiles() method to replace mimemsg.
#           Use NewsTool.messageDate() for timestamp.
#           Use string methods versus module.
#
#   2.2.23  2003-07-01  Todd Valentic
#           Fixed problem with initial history roll list.
#           Added support for .zip, .gz compression (in addition to .bz2)
#           Added check if source and output filenames are same (do nothing)
#
#   2.2.24  2003-08-14  Todd Valentic
#           createCallback -> getCallback
#
#   2.2.25  2003-08-21  Todd Valentic
#           Make sure to close files before calling the callback.
#
#   2.1.26  2003-08-23  Todd Valentic
#           Added option (output.overwrite) to control whether the output
#               file is overwritten if it already exists. Defaults to true.
#           Use get_timedelta for timeout and pollrate.
#
#   2.1.27  2003-10-03  Todd Valentic
#           Use get_timedelta for input.timeout and input.timegap
#
#   2.1.28  2004-03-04  Todd Valentic
#           Added timespan feature. We can now start a new file based
#               a span of time (ie daily) when concatenating messages.
#               It works in a manner similar to timegap and timeout.
#
#   2.1.29  2004-04-08  Todd Valentic
#           Add wait_on_start option (defaults on) to pause on startup
#               (prevents a lot of thrashing the news server when all
#               the processes are starting), but for long waits, it can
#               be annoying for debugging.
#
#           Added start_current_reset (defaults on) to prevent reseting
#               the start_current counter, allowing you to "catchup"
#               each time the news server is queried. Therefore, you
#               can now do something like wake up each hour and just
#               process the last message to give you an hourly sample.
#
#   2.1.30  2004-08-17  Todd Valentic
#           Change to new logging interface.
#
#   2.1.31  2004-12-27  Todd Valentic
#           Convert from mx.DateTime to datetime
#
#   2.1.32  2005-01-10  Todd Valentic
#           Make sure all times are UTC. There were places we were
#               comparing local times (which is usually UTC on the
#               servers anyways) with message headers. The new
#               datetime object is smart enough to know when you are
#               trying to compare an object with a timezone and one
#               without. Note that calling utcnow() returns the time
#               in UTC, but does not set the timezone. Use now(utc)
#               instead.
#
#           Made the report and summary enables default to the enable
#               state of the newsposter object.
#
#   2.1.33  2005-03-23  Todd Valentic
#           Fixed bug checking time.
#
#   2.1.34  2005-06-15  Todd Valentic
#           Added message headers in pattern replacement on filenames.
#
#   2.1.35  2005-08-12  Todd Valentic
#           Modified timespan to work with max_messages. By setting
#               both, we can control how many messages go into each
#               span. This can be used to do things like saving a
#               single webcam image taken once a minute every hour.
#
#   2.1.36  2005-08-13  Todd Valentic
#           Use parser for interpreting start/stop times.
#
#   2.1.37  2005-09-19  Todd Valentic
#           Added hack to save message time for callbacks. Need to
#               come up with a better method at some point.
#
#   2.1.38  2006-07-18  Todd Valentic
#           Added save_handler to allow for other saving options
#               besides copy/append (i.e., save to HDF).
#           Improve exit on error handling
#
#   2.1.39  2006-09-30  Todd Valentic
#           Fixed bug in timegap reporting. Don't reset checkpoint
#               if timegap and max_messages set.
#
#   2.1.40  2007-07-18  Todd Valentic
#           Added modify_message() call in process_message to allow
#               inherited classes to manipulate the message before
#               processing. One common use is for injecting new
#               headers (ie message times).
#
#   2.1.41  2008-05-13  Todd Valentic
#           Added parts offset
#
#   2.1.42  2008-07-19  Todd Valentic
#           Rewrote the split message processing to make rejoins more
#               robust. No longer need the parts offset added in
#               2.1.41. Removing.
#           Post errors to trouble newsgroup.
#           Make sure that keep_summary is initialized in Checkpoint.
#
#   2.1.43  2008-11-133 Todd Valentic
#           Add get_file_time() method to allow the output filename
#               computation to be based on the a timestamp relevant
#               to the file (instead of the message). The use case is
#               for messages with multiple attached files, like we
#               have in the RUDICS transport.
#
#   2.1.44  2009-02-04  Todd Valentic
#           Add newsgroup as parameter to get_file_time()
#           Print traceback if problem parsing output.rules()
#
#   2.1.45  2009-05-11  Todd Valentic
#           md5 module deprecated in Python 2.6. Use a import check
#               to pull in right library on older systems.
#
#   2.1.26  2009-09-24  Todd Valentic
#           Encounterd problem with strftime and very long (127+)
#               format strings. These are common in deep directory
#               heirarchies. Temp fix is to move it after the other
#               expansions, but that is only of limited help. Need
#               to find a work around...
#
#           Only update checkpoint message count on good files (it was
#               accounting for partial files when they are split over
#               multiple messages.
#
#   2.1.27  2010-05-07  Todd Valentic
#           Reset message count if low_mark > currentMessage. This
#               state can happen if the newsgroup is moved to a new
#               server and the counts are reset.
#
#   2.1.28  2012-02-28  Todd Valentic
#           Add file check before trying to set permissions (allows
#               for destname=/dev/null or alternate save_handlers
#               that don't actually save a file).
#
#   2.1.19  2012-11-13  Todd Valentic
#           Check for existance of database in __del__()
#
#   2.1.20  2013-01-27  Todd Valentic
#           Fix typo in __del__ for database check. Change to try..except
#
#   2.1.21  2013-05-24  Todd Valentic
#           Add newsgroup name into working directory for chunked messages
#
#   2.1.22  2018-04-16  Todd Valentic
#           Add validate_file
#
#   2.1.23  2021-07-23  Todd Valentic
#           Add ability to specify news server port (Default 119)
#
#   2016-12-23  Todd Valentic
#               Remove python<2.7 md5 import
# 		        Use datetime.timezone.utc for utc
#               No longer using dateutil.parser
#               Add main()
#
#   2022-10-07  Todd Valentic
#               Updates for Python3
#                   commands -> subprocess
#                   removeFile -> remove_file
#                   sizeDesc -> size_desc
#                   getInt -> get_int
#                   getDeltaTime -> get_timedelta
#                   getboolean -> get_boolean
#                   NewsPoster
#
#   2023-07-26  Todd Valentic
#               Updated for transport3 / python3
#
##############################################################################

import email
import fnmatch
import glob
import math
import os
import shelve
import shutil
import stat
import sys
import subprocess
import traceback

from datetime import datetime, timezone
from hashlib import md5
from pathlib import Path

from datatransport import ProcessClient
from datatransport import NewsPoster
from datatransport import newstool

from datatransport.utilities import size_desc
from datatransport.utilities import remove_file
from datatransport.utilities import PatternTemplate
from datatransport.utilities import datefunc
from datatransport.utilities import make_path


def now():
    """Return current time"""
    return datetime.now(timezone.utc)


class Checkpoint:
    """Track newsgroup state"""

    def __init__(self, newsgroup, low_mark, max_history, keep_summary):
        self.newsgroup = newsgroup
        self.last_message = low_mark - 1

        self.keep_summary = keep_summary
        self.summary_max_files = 1000
        self.reset_summary()

        self.num_messages = 0
        self.num_bytes = 0
        self.filenames = {}

        self.last_time = now()
        self.start_time = now()
        self.poll_time = now()

        self.history = []
        self.max_history = max_history

    def update_message(self, msg_num, date=None):
        """Update current state"""

        self.num_messages = self.num_messages + 1
        self.last_message = msg_num
        self.poll_time = now()

        if date:
            self.last_time = date

    def touch_message(self, msg_num):
        """Mark message time"""

        self.last_message = msg_num
        self.poll_time = now()

    def update_file(self, filename, numbytes):
        """Update file stats"""

        if filename in self.filenames:
            (prev_msgs, prev_bytes) = self.filenames[filename]
        else:
            prev_msgs = 0
            prev_bytes = 0

        self.filenames[filename] = (prev_msgs + 1, prev_bytes + numbytes)
        self.num_bytes = self.num_bytes + numbytes

    def reset(self, date=None, last_message=0):
        """Reset state"""

        if not date:
            date = now()

        if self.filenames:
            self.history.insert(0, (self.filenames, self.last_time))
            self.history = self.history[0 : self.max_history + 1]

            if self.keep_summary:
                self.summary_files.append(self.filenames)
                self.summary_files = self.summary_files[-self.summary_max_files :]
                self.summary_bytes = self.summary_bytes + self.num_bytes
                self.summary_msgs = self.summary_msgs + self.num_messages

        self.start_time = date
        self.last_time = date
        self.num_messages = 0
        self.num_bytes = 0
        self.filenames = {}
        self.last_message = last_message

    def is_active(self):
        """Is newsgroup active"""

        return self.num_messages > 0

    def in_span(self, curtime, timespan):
        """Check if time in span"""

        lastsecs = self.last_time.timestamp()
        cursecs = curtime.timestamp()
        interval = timespan.total_seconds()

        lastblock = int(math.floor(lastsecs / interval))
        curblock = int(math.floor(cursecs / interval))

        return lastblock == curblock

    def reset_summary(self):
        """Reset summary counts"""

        self.summary_files = []
        self.summary_start = now()
        self.summary_bytes = 0
        self.summary_msgs = 0

    def get_report(self, title, reason):
        """Generate report"""

        message = []

        fmt = "%c %Z %z"

        message.append("_" * 75)
        message.append(title)
        message.append("")
        message.append(f"    Newsgroup     : {self.newsgroup}")
        message.append(f"    Beginning     : {self.start_time.strftime(fmt)}")
        message.append(f"    Ending        : {self.last_time.strftime(fmt)}")
        message.append(f"    Messages      : {self.num_messages}")
        message.append(f"    Total bytes   : {size_desc(self.num_bytes)}")
        message.append(f"    Reason        : {reason}")
        message.append("")

        if self.filenames:
            message.append("    Filenames")
            message.append("    ---------")

            for filename in sorted(self.filenames):
                msgs, numbytes = self.filenames[filename]
                if msgs > 1:
                    units = "msgs"
                else:
                    units = "msg"
                message.append(
                    f"    {filename}: {size_desc(numbytes)} ({msgs} {units})"
                )

        message.append("_" * 75)
        message = "\n".join(message)

        return message

    def get_summary(self, title):
        """Get summary report"""

        message = []

        fmt = "%c %Z %z"

        message.append("_" * 75)
        message.append(title)
        message.append("")
        message.append(f"    Newsgroup     : {self.newsgroup}")
        message.append(f"    Beginning     : {self.summary_start.strftime(fmt)}")
        message.append(f"    Ending        : {now().strftime(fmt)}")
        message.append(f"    Messages      : {self.summary_msgs}")
        message.append(f"    Total bytes   : {size_desc(self.summary_bytes)}")
        message.append("")

        if self.summary_files:
            message.append("    Filenames")
            message.append("    ---------")

            for filegroup in self.summary_files:
                prefix = "-"
                for filename in sorted(filegroup):
                    msgs, numbytes = filegroup[filename]
                    if msgs > 1:
                        units = "msgs"
                    else:
                        units = "msg"
                    message.append(
                        f"  {prefix} {filename}: {size_desc(numbytes)} ({msgs} {units})"
                    )
                    prefix = " "

        message.append("_" * 75)
        message = "\n".join(message)

        return message


class ArchiveGroups(ProcessClient):
    """Process Client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.report_poster = NewsPoster(self, prefix="report", quiet=True)
        self.summary_poster = NewsPoster(self, prefix="summary", quiet=True)
        self.trouble_poster = NewsPoster(self, prefix="trouble", quiet=True)

        self.time_fmt = "%Y %j %H:%M:%S"

        self.replace_proc_group = PatternTemplate("procgroup", "/")
        self.replace_client_name = PatternTemplate("clientname", "/")
        self.replace_newsgroup = PatternTemplate("newsgroup", ".")
        self.replace_filename = PatternTemplate("filename", ".")
        self.replace_rule = PatternTemplate("rule")
        self.replace_history = PatternTemplate("history")
        self.replace_header = PatternTemplate("header")

        self.time_offset = self.config.get_timedelta("timeoffset", 0)
        self.pollrate = self.config.get_rate("input.pollrate", 60)
        self.wait_on_start = self.config.get_boolean("input.wait_on_start", True)

        self.pollserver_host = self.config.get("input.server", "localhost")
        self.pollserver_port = self.config.get_int("input.server.port", 119)
        self.timeout = self.config.get_timedelta("input.timeout", None)
        self.timegap = self.config.get_timedelta("input.timegap", None)
        self.timespan = self.config.get_timedelta("input.timespan", None)
        self.start_time = self.config.get("input.start_time", None)
        self.stop_time = self.config.get("input.stop_time", None)
        self.start_current = self.config.get_int("input.start_current", 1)
        self.start_current_reset = self.config.get_boolean(
            "input.start_current.reset", True
        )

        self.dest_path = self.config.get("output.path", "")
        self.dest_name = self.config.get("output.name", "<rule>")
        self.dest_file_mode = self.config.get_int("output.mode.file", "0o644")
        self.dest_path_mode = self.config.get_int("output.mode.path", "0o755")
        self.max_messages = self.config.get_int("output.max_messages", 1)
        self.expire = self.config.get_timedelta("output.expire", None)
        self.history = self.config.get_int("output.history", 0)
        self.plain_text_name = self.config.get("output.plain_text_name", "noname.txt")
        self.uncompress = self.config.get_boolean("output.uncompress", False)
        self.overwrite = self.config.get_boolean("output.overwrite", True)

        self.report_subject = self.config.get(
            "report.subject", "Newsgroup archive report"
        )
        self.report_title = self.config.get("report.title", "Newsgroup archive report")

        enable = self.report_poster.enabled
        self.report_max_messages = self.config.get_boolean(
            "report.enable.max_messages", enable
        )
        self.report_timeout = self.config.get_boolean("report.enable.timeout", enable)
        self.report_timegap = self.config.get_boolean("report.enable.timegap", enable)
        self.report_timespan = self.config.get_boolean("report.enable.timespan", enable)

        self.summary_subject = self.config.get(
            "summary.subject", "Newsgroup summary report"
        )
        self.summary_title = self.config.get(
            "summary.title", "Newsgroup summary report"
        )
        self.summary_enable = self.config.get_boolean(
            "summary.enable", self.summary_poster.enabled
        )
        self.summary_period = self.config.get_timedelta("summary.period", 60 * 60 * 24)
        self.summary_offset = self.config.get_timedelta("summary.offset", 0)
        self.summary_max_files = self.config.get_int("summary.max_files", 1000)

        self.debug = self.config.get_boolean("debug", False)

        self.include_newsgroups = self.config.get_list("input.newsgroups")
        self.exclude_newsgroups = self.config.get_list("input.newsgroups.exclude")
        self.include_filenames = self.config.get_list("input.filenames", "*")
        self.exclude_filenames = self.config.get_list("input.filenames.exclude")

        try:
            self.callback = self.config.get_callback("callback")
        except:
            self.log.exception("Problem loading callback")
            self.abort()

        try:
            self.rules = eval(
                self.config.get("output.rules", "[('*','*','<filename>')]")
            )
        except:
            self.log.exception("There was a problem parsing output.rules.")
            self.abort()

        self.validate_setup()
        self.setup_database()

    def __del__(self):
        try:
            self.database.close()
        except:
            pass

    def setup_database(self):
        """Setup checkpoint database"""

        self.database = shelve.open("checkpoint.db")

        # Update the config parameters in the database to new values.

        for key, checkpoint in self.database.items():
            checkpoint.max_history = self.history
            checkpoint.keep_summary = self.summary_enable
            checkpoint.summary_max_files = self.summary_max_files
            self.database[key] = checkpoint

        # Make sure that the on-disk files are initialized, otherwise they will
        # be zero-length and cause problems next time if opened that way.

        self.database.sync()

    def validate_setup(self):
        """Validate setup"""

        if self.start_time:
            try:
                self.start_time = datefunc.strptime(
                    self.start_time, self.time_fmt, timezone.utc
                )
                self.log.info("Only processing messages past %s", self.start_time)
            except:
                self.log.exception(
                    "The start time format is bad. It should be YYYY JJJ HH:MM:SS"
                )
                self.abort("Exiting")

        if self.stop_time:
            try:
                self.stop_time = datefunc.strptime(
                    self.stop_time, self.time_fmt, timezone.utc
                )
                self.log.info("Only processing messages before  %s", self.stop_time)
            except:
                self.log.exception(
                    "The stop time format is bad. It should be YYYY JJJ HH:MM:SS"
                )
                self.abort("Exiting")

        self.log.debug("Watching these newsgroups (+ = include, - = exclude):")
        for newsgroup in self.include_newsgroups:
            self.log.debug("  + %s", newsgroup)
        for newsgroup in self.exclude_newsgroups:
            self.log.debug("  - %s", newsgroup)

        self.log.debug("Watching these filenames (+ = include, - = exclude):")
        for filename in self.include_filenames:
            self.log.debug("  + %s", filename)
        for filename in self.exclude_filenames:
            self.log.debug("  - %s", filename)

    def get_groups(self, newsserver):
        """Get newsgroup list"""

        groups = []

        _response, newsgroup_list = newsserver.list()

        for newsgroup, high_mark, low_mark, _flag in newsgroup_list:
            high_mark = int(high_mark)
            low_mark = int(low_mark)

            for incgroup in self.include_newsgroups:
                if fnmatch.fnmatch(newsgroup, incgroup):
                    keep = 1
                    for excgroup in self.exclude_newsgroups:
                        if fnmatch.fnmatch(newsgroup, excgroup):
                            keep = 0
                            break
                    if keep:
                        groups.append((newsgroup, low_mark, high_mark))
                        if not newsgroup in self.database:
                            checkpoint = Checkpoint(
                                newsgroup, low_mark, self.history, self.summary_enable
                            )
                            self.database[newsgroup] = checkpoint
                            self.database.sync()

        return groups

    def filter_filenames(self, filenames):
        """Filter filenames"""

        goodfiles = []

        for filename in filenames:
            for incpattern in self.include_filenames:
                if fnmatch.fnmatch(filename, incpattern):
                    keep = 1
                    for excpattern in self.exclude_filenames:
                        if fnmatch.fnmatch(filename, excpattern):
                            keep = 0
                            break
                    if keep:
                        goodfiles.append(filename)

        return goodfiles

    def get_file_time(self, _newsgroup, _filename):
        """Get timestamp from filename"""
        return None

    def compute_name(self, filename, checkpoint):
        """Compute filename"""

        rule = self.find_rule(checkpoint.newsgroup, filename)

        path = os.path.join(self.dest_path, self.dest_name)
        path = self.replace_rule(path, rule)

        try:
            filetime = self.config.get_file_time(checkpoint.newsgroup, filename)
        except:
            filetime = None

        if filetime is None:
            if self.max_messages == 1:
                filetime = checkpoint.last_time
            else:
                filetime = checkpoint.start_time

        # NOTE - Potential breakage with long paths and versions
        #        of python < 2.4.4 where strftime is limited to
        #        127 character strings. Moving it to the end
        #        of the expansion here, helped, but I need to
        #        figure out a better way of doing this...

        path = self.replace_newsgroup(path, checkpoint.newsgroup)
        path = self.replace_proc_group(path, self.groupname)
        path = self.replace_client_name(path, self.name)
        path = self.replace_filename(path, str(filename))
        path = self.replace_header(path)
        path = filetime.strftime(path)

        return path

    def find_rule(self, newsgroup, filename):
        """Find matching rule"""

        for rule in self.rules:
            if fnmatch.fnmatch(newsgroup, rule[0]) and fnmatch.fnmatch(
                filename, rule[1]
            ):
                return rule[2]

        return ""

    def post_report(self, checkpoint, reason):
        """Post report"""

        title = self.report_title
        title = self.replace_newsgroup(title, checkpoint.newsgroup)
        title = self.replace_proc_group(title, self.groupname)
        title = self.replace_client_name(title, self.name)

        header = self.report_subject
        header = self.replace_newsgroup(header, checkpoint.newsgroup)
        header = self.replace_proc_group(header, self.groupname)
        header = self.replace_client_name(header, self.name)

        message = checkpoint.get_report(title, reason)

        try:
            self.report_poster.set_subject(header)
            self.report_poster.post_text(message)
            self.log.info("%s, posting report (%s)", reason, checkpoint.newsgroup)
        except:
            self.log.exception("Error posting report message")

    def post_summary(self, checkpoint):
        """Post summary report"""

        title = self.summary_title
        title = self.replace_newsgroup(title, checkpoint.newsgroup)
        title = self.replace_proc_group(title, self.groupname)
        title = self.replace_client_name(title, self.name)

        header = self.summary_subject
        header = self.replace_newsgroup(header, checkpoint.newsgroup)
        header = self.replace_proc_group(header, self.groupname)
        header = self.replace_client_name(header, self.name)

        message = checkpoint.get_summary(title)

        try:
            self.summary_poster.set_subject(header)
            self.summary_poster.post_text(message)
            self.log.info("Posting summary report (%s)", checkpoint.newsgroup)
        except:
            self.log.error("Error posting summary message")

    def roll_files(self, checkpoint):
        """Only keep N files"""

        self.log.debug("Rolling file history:")
        self.log.debug("  history length: %d", len(checkpoint.history))
        self.log.debug("  max history: %d", checkpoint.max_history)

        if not checkpoint.history or checkpoint.max_history == 0:
            self.log.debug("  no files to roll")
            return

        history_index = list(range(len(checkpoint.history)))
        history_index.reverse()

        for index in history_index:
            filenames = checkpoint.history[index][0].keys()

            for filename in filenames:
                oldname = self.replace_history(filename, index)
                newname = self.replace_history(filename, index + 1)

                try:
                    if index < checkpoint.max_history:
                        if oldname == newname:
                            self.log.debug("  keeping  %s", oldname)
                        else:
                            os.rename(oldname, newname)
                            self.log.debug("  copying  %s -> %s", oldname, newname)
                    else:
                        os.remove(oldname)
                        self.log.debug("  removing %s", oldname)
                except:
                    pass

    def save_callback(self, filename):
        """Save callback"""

        # Used by derived classes to do something interesting with
        # the newly saved file such as creating a thumbnail of images.

    def save_handler(self, srcname, destname, mode):
        """Save handler"""

        srcfile = open(srcname, "rb")
        destfile = open(destname, mode)
        shutil.copyfileobj(srcfile, destfile)
        srcfile.close()
        destfile.close()
        os.chmod(destname, self.dest_file_mode)

    def save_file(self, srcname, checkpoint):
        """Save single file"""

        num_bytes = os.stat(srcname)[stat.ST_SIZE]
        name = self.compute_name(srcname, checkpoint)
        destname = self.replace_history(name, str(0))
        destname = os.path.abspath(destname)

        self.log.debug("Saving %s", srcname)
        self.log.debug("  dest: %s", destname)
        self.log.debug("  maxmessages: %d", self.max_messages)
        self.log.debug("  nummessages: %d", checkpoint.num_messages)
        self.log.debug("  is_active: %s", checkpoint.is_active())

        if not self.overwrite and os.path.exists(destname):
            self.log.debug("  destination exists - not saving")
            return

        if self.max_messages != 1 and checkpoint.is_active():
            mode = "ab"
            self.log.info("Concat: %s -> %s", srcname, destname)
        else:
            mode = "wb"
            self.log.info("Start : %s -> %s", srcname, destname)

        try:
            make_path(destname, self.dest_path_mode)
        except BaseException as err:
            self.log.exception("Problem creating the path.")
            raise err

        try:
            if os.path.abspath(srcname) != destname:
                self.save_handler(srcname, destname, mode)

            if os.path.isfile(destname):
                os.chmod(destname, self.dest_file_mode)
        except:
            self.log.exception("Problem saving file")
            raise

        try:
            self.save_callback(destname)
        except:
            self.log.exception("Problem running the save callback")

        checkpoint.update_file(name, num_bytes)

    def uncompress_file(self, cmd, filename):
        """Uncompress file if needed"""

        uncompressname = filename.stem

        try:
            status, output = subprocess.getstatusoutput(f"{cmd} {filename}")
        except IOError:
            self.abort("Error running uncompression: %s", cmd)

        if status == 0:
            self.log.debug("  uncompressing: %s", filename)
        else:
            subject = "Archive: Error umcompressing file"
            note = []
            note.append("Error trying to uncompress file")
            note.append("   command: %s", cmd)
            note.append("   status:  %d", status)
            note.append("   output:  %s", output)
            self.post_error(subject, note)
            return None

        return Path(uncompressname)

    def last_tracback(self):
        """Format last traceback"""

        return traceback.format_tb(sys.exc_traceback)[-1]

    def post_error(self, subject, note):
        """Post error message"""

        for line in note:
            self.log.error(line)

        try:
            self.trouble_poster.set_subject(subject)
            self.trouble_poster.post_text("\n".join(note))
            self.log.info("Sent trouble message")
        except:
            self.log.exception("Failed to post trouble message")

    def validate_file(self, _filename):
        """Validate file"""
        return True

    def process_files(self, filenames, checkpoint, msg_num, msg_time):
        """Process files"""

        # Hack alert - the callbacks sometimes want access to
        # the message time.
        self.msg_time = msg_time

        # Roll old files

        if not checkpoint.is_active():
            self.roll_files(checkpoint)
            checkpoint.reset(msg_time, msg_num)

        # Uncompress any files that need it

        uncompressedfiles = []

        for filename in filenames:
            if self.uncompress:
                # if filename.suffix == ".zip":
                #    filename = self.uncompress_file("unzip -o", filename)
                if filename.suffix == ".gz":
                    filename = self.uncompress_file("gunzip -f", filename)
                elif filename.suffix == ".bz2":
                    filename = self.uncompress_file("bunzip2 -f", filename)

            if filename:
                uncompressedfiles.append(filename)

        # Run the callback function if needed

        savefiles = []

        for filename in uncompressedfiles:
            if not self.validate_file(filename):
                continue

            if self.callback:
                try:
                    filename = self.callback(self, filename)
                    # filename can now be the same name, a new name
                    # None if we should forget it, or a list if new
                    # files were created.
                except:
                    self.log.exception("Problem running callback:")
                    continue

            if not filename:
                # The callback doesn't want the file saved
                continue

            if isinstance(filename, (list, tuple)):
                savefiles.extend(filename)
            else:
                savefiles.append(filename)

        for filename in savefiles:
            try:
                self.save_file(filename, checkpoint)
            except:
                note = []
                note.append("Problem storing file:")
                note.append(f"  filename    = {filename}")
                note.append(f"  path        = {self.dest_path}")
                note.append(f"  message num = {msg_num}")
                note.append(f"  newsgroup   = {checkpoint.newsgroup}")
                note.append(f"  file mode   = {oct(self.dest_file_mode)}")
                note.append(f"  path mode   = {oct(self.dest_path_mode)}")
                note.append("")
                note.append(traceback.format_exc())

                subject = f"Archive: problem saving {filename}"

                self.post_error(subject, note)

                if self.debug:
                    self.log.error("Exiting on error (debug mode on)")
                    raise

                self.abort("Exiting on error")
                break

            remove_file(filename)

            if self.is_stopped():
                break

    def modify_message(self, message):
        """Modify message"""

        return message

    def process_message(self, newsserver, msg_num, checkpoint):
        """Process message"""

        try:
            article = newsserver.article(str(msg_num))[1]
            message = email.message_from_bytes(b"\n".join(article.lines))
        except:
            self.log.exception("    unable to retrieve article body")
            checkpoint.touch_message(msg_num)
            return

        message = self.modify_message(message)

        if not message:
            self.log.info("    skipping message")
            checkpoint.touch_message(msg_num)
            return

        msg_time = newstool.message_date(message) + self.time_offset

        headers = {}
        for key in message.keys():
            headers[key] = message[key]

        self.replace_header.set_value(headers)

        if self.start_time and msg_time < self.start_time:
            self.log.debug(
                "    rejecting - too old (%s)", msg_time.strftime(self.time_fmt)
            )
            checkpoint.reset(msg_time, msg_num)
            return

        if self.stop_time and msg_time > self.stop_time:
            self.log.debug(
                "    rejecting - too new (%s)", msg_time.strftime(self.time_fmt)
            )
            checkpoint.reset(msg_time, msg_num)
            return

        if self.timegap and (msg_time - checkpoint.last_time) >= self.timegap:
            self.log.info("    time gap between messages exceeds limit")
            if self.report_timegap and checkpoint.is_active():
                self.post_report(checkpoint, "Time gap")
            checkpoint.reset(msg_time, msg_num)

        if self.timespan:
            if checkpoint.in_span(msg_time, self.timespan):
                if self.max_messages and checkpoint.num_messages >= self.max_messages:
                    self.log.debug("    rejecting, max reached for this time span")
                    return

            else:
                self.log.info("    message is in a new time span")
                if self.report_timespan and checkpoint.is_active():
                    self.post_report(checkpoint, "Time span")
                checkpoint.reset(msg_time, msg_num)

        try:
            filenames = newstool.save_files(message)
        except:
            self.log.exception("    rejecting - unable to parse message body")
            checkpoint.touch_message(msg_num)
            return

        if "x-transport-part" in message:
            # Assumes that there is only one part attached

            part, total = map(int, message["X-Transport-Part"].split("/"))
            srcname = os.path.basename(filenames[0])
            destname = os.path.basename(message["X-Transport-Filename"])
            workdir = os.path.join("work", checkpoint.newsgroup, destname)
            workname = os.path.join(workdir, srcname)

            if not os.path.exists(workdir):
                os.makedirs(workdir)

            shutil.move(srcname, workname)

            parts = sorted(glob.glob(os.path.join(workdir, "*")))

            if len(parts) == total:
                with Path(destname).open("wb") as destfile:
                    for filepart in parts:
                        contents = Path(filepart).read_bytes()
                        destfile.write(contents)

                remove_file(parts)
                os.rmdir(workdir)

                self.log.debug("Joined split message: %d parts", total)

                if "x-transport-md5" in message:
                    orgmd5 = message["x-transport-md5"]
                    newmd5 = md5(Path(destname).read_bytes()).hexdigest()
                    if orgmd5 == newmd5:
                        self.log.debug("  MD5 checksum matches")
                        filenames = [destname]
                    else:
                        subject = "Archive: MD5 mismatch"

                        note = []
                        note.append("MD5 mistmatch in file")
                        note.append(f"  message ID: {message['Xref']}")
                        note.append(f"  filename:   {destname}")
                        note.append(f"  expected:   {orgmd5}")
                        note.append(f"  found:      {newmd5}")
                        self.post_error(subject, note)

                        filenames = []
                else:
                    self.log.debug("  No MD5 checksum to compare")
                    filenames = [destname]

            else:
                filenames = []
                self.log.debug("Joining split message: part %d of %d", part, total)

        goodfiles = self.filter_filenames(filenames)
        self.log.debug(
            "    message contains %d file(s), saving %d:",
            len(filenames),
            len(goodfiles),
        )
        for filename in filenames:
            if filename in goodfiles:
                self.log.debug("      + %s", filename)
            else:
                self.log.debug("      - %s", filename)

        if goodfiles:
            self.process_files(goodfiles, checkpoint, msg_num, msg_time)
            checkpoint.update_message(msg_num, msg_time)

        remove_file(filenames)

        self.log.debug(
            "    after writing files, message count=%d, max=%d",
            checkpoint.num_messages,
            self.max_messages,
        )

        if self.max_messages and checkpoint.num_messages >= self.max_messages:
            if self.max_messages > 1:
                self.log.info("    max message count reached")
            if self.report_max_messages:
                self.post_report(checkpoint, "Max message count reached")
            if not self.timespan and not self.timegap:
                checkpoint.reset(msg_time, msg_num)

    def process_group(self, newsserver, checkpoint, low_mark, high_mark):
        """Process newsgroup"""

        self.log.debug("Processing newsgroup: %s", checkpoint.newsgroup)
        self.log.debug(
            "  available message numbers: %d - %d, last processed: %d",
            low_mark,
            high_mark,
            checkpoint.last_message,
        )

        if self.start_current == 1:
            self.log.debug("  catching up")
            checkpoint.reset(last_message=high_mark)

        elif self.start_current < 0:
            last_msg = max(low_mark - 1, high_mark + self.start_current)
            last_msg = max(last_msg, checkpoint.last_message)
            self.log.debug(
                "  starting with last %d message(s)", abs(high_mark - last_msg)
            )
            checkpoint.reset(last_message=last_msg)

        elif low_mark > high_mark:
            self.log.debug("  no messages on server")
            return

        elif (
            checkpoint.last_message < low_mark - 1
            or checkpoint.last_message > high_mark
        ):
            self.log.info("  message count reset")
            checkpoint.reset(last_message=low_mark - 1)

        if checkpoint.last_message == high_mark:
            self.log.debug("  no new messages")

            if self.timeout and checkpoint.is_active():
                if (now() - checkpoint.poll_time) >= self.timeout:
                    self.log.info("  time out reached")
                    if self.report_timeout:
                        self.post_report(checkpoint, "Time out reached")
                    checkpoint.reset(last_message=high_mark)

                return

        newsserver.group(checkpoint.newsgroup)

        for msg_num in range(checkpoint.last_message + 1, high_mark + 1):
            self.log.debug("  processing message %d", msg_num)

            self.process_message(newsserver, msg_num, checkpoint)
            self.database[checkpoint.newsgroup] = checkpoint
            self.database.sync()

            if self.is_stopped():
                break

    def check_groups(self):
        """Check newsgroups"""

        host = self.pollserver_host
        port = self.pollserver_port

        with newstool.NewsTool().open_server(host, port) as newsserver:
            for newsgroup, low_mark, high_mark in self.get_groups(newsserver):
                checkpoint = self.database[newsgroup]
                self.process_group(newsserver, checkpoint, low_mark, high_mark)
                self.database[newsgroup] = checkpoint
                self.database.sync()

                if self.is_stopped():
                    break

        if self.start_current_reset:
            self.start_current = 0

    def expire_files(self):
        """Expire old files"""

        self.log.debug("Expiring old files")

        for key, checkpoint in self.database.items():
            self.log.debug("  - %s", key)

            if len(checkpoint.history) > 0:
                self.log.debug("    checking file history expire times:")

                history_index = list(range(len(checkpoint.history)))
                history_index.reverse()

                for index in history_index:
                    filenames, lastdate = checkpoint.history[index]

                    diff = now() - lastdate
                    expire = diff > self.expire

                    self.log.debug(
                        "      %d: time=%s, diff=%s, expire? %d",
                        index,
                        lastdate,
                        diff,
                        expire,
                    )

                    if expire:
                        for filename in filenames.keys():
                            filename = self.replace_history(filename, index)
                            self.log.debug("         %s", filename)
                            try:
                                os.remove(filename)
                            except:
                                pass
                        del checkpoint.history[index]
                    else:
                        break

            self.database[key] = checkpoint

        self.database.sync()

    def check_summary(self):
        """Check if time for summary report"""

        if not self.summary_enable:
            return

        self.log.debug("Checking summary reporting")

        for key, checkpoint in self.database.items():
            self.log.debug("  - %s", key)

            secs = checkpoint.summary_start.timestamp()
            interval = self.summary_period.total_seconds()
            offset = self.summary_offset.total_seconds()

            next_time = (math.floor(secs / interval) + 1) * interval + offset
            next_time = datetime.fromtimestamp(next_time, tz=timezone.utc)

            self.log.debug("    - interval = %s", self.summary_period)
            self.log.debug("    - offset   = %s", self.summary_offset)
            self.log.debug("    - start_time= %s", checkpoint.summary_start)
            self.log.debug("    - next_time = %s", next_time)

            if now() >= next_time:
                self.log.info("Summary time reached")
                self.post_summary(checkpoint)
                checkpoint.reset_summary()

            self.database[key] = checkpoint

        self.database.sync()

    def main(self):
        """Main application"""

        self.log.info("Starting")

        while self.wait(self.pollrate):
            try:
                self.check_groups()
            except SystemExit:
                self.log.info("TAV *** SYSTEM EXIT exception")
                break
            except:
                note = [traceback.format_exc()]
                self.post_error("Archive: problem checking groups", note)

            self.check_summary()

            if self.expire:
                self.expire_files()

        self.log.info("Finished")


def main():
    """Script entry point"""
    ArchiveGroups(sys.argv).run()



#!/usr/bin/env python3

##########################################################################
#
#   MessageBox Posting example
#
#   Periodically post a message to a messagebox 
#
#   2023-02-22  Todd Valentic
#               Initial implementation
#
##########################################################################

from datatransport import ProcessClient
from datatransport import MessageBoxPoster 

import sys

class Client (ProcessClient):

    def __init__(self, args):
        ProcessClient.__init__(self, args)

        self.poster = MessageBoxPoster(self)

        self.rate = self.get_rate('rate')
        self.text = self.get('text')

        self.log.info(f'Post message every {self.rate.period}')

    def run(self):

        self.log.info('Start')

        while self.wait(self.rate):
            self.poster.post_text(self.text)
            self.log.info('Posted message')

        self.log.info('Stop')

if __name__ == '__main__':
    Client(sys.argv).run()




#!/usr/bin/env python3
"""Hello World example"""

##########################################################################
#
#   Hello World example
#
#   Periodically print a message to the log file.
#
#   2022-10-07  Todd Valentic
#               Initial implementation
#
##########################################################################

import sys

from datatransport import ProcessClient


class Client(ProcessClient):
    """Process Client"""

    def __init__(self, args):
        ProcessClient.__init__(self, args)

        self.rate = self.config.get_rate("rate")
        self.text = self.config.get("text")

        self.log.info("Log message every: %s", self.rate.period)

    def main(self):
        """Main application"""

        while self.wait(self.rate):
            self.log.info(self.text)


if __name__ == "__main__":
    Client(sys.argv).run()



#!/usr/bin/env python3
"""Directory Service"""

##########################################################################
#
#   Directory Service
#
#   The directory is used to return information about the
#   various services. It is used primarily in bootstraping
#   by client programs. Based on AMISR directory service.
#
#   2005-11-08  Todd Valentic
#               Initial implementation.
#
#   2006-06-14  Todd Valentic
#               Add label to XMLRPCServerMixin call
#
#   2006-09-16  Todd Valentic
#               Added hold to Service.
#
#   2007-08-19  Todd Valentic
#               Allow directory port to be set.
#
#   2013-01-30  Todd Valentic
#               Make sure port has a value.
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#
#   2023-07-27  Todd Valentic
#               Updated for transport3 / python3
#
#   2023-09-06  Todd Valentic
#               Return dict in list
#
##########################################################################

import sys

from datatransport import ProcessClient
from datatransport import ConfigComponent
from datatransport import XMLRPCServer


class Service(ConfigComponent):
    """Service Component"""

    def __init__(self, name, config, parent, **kw):
        ConfigComponent.__init__(self, "service", name, config, parent, **kw)

        self.host = self.config.get("host", "localhost")
        self.port = self.config.get_int("port")
        self.scheme = self.config.get("scheme", "http")
        self.label = self.config.get("label")
        self.url = f"{self.scheme}://{self.host}:{self.port}"
        self.hold = self.config.get_boolean("hold", True)

        if not self.port:
            raise ValueError("No port specified")

    def info(self):
        
        return {
            "host": self.host,
            "port": self.port,
            "scheme": self.scheme,
            "label": self.label,
            "url": self.url,
            "hold": self.hold
        }

class DirectoryService(ProcessClient):
    """Process Client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        port = self.config.get_int("directory.port", 8411)

        self.xmlserver = XMLRPCServer(
            self, queue_size=100, port=port, label="Directory"
        )
        self.main = self.xmlserver.main

        self.xmlserver.register_function(self.list)
        self.xmlserver.register_function(self.lookup, "get")

        self.services = self.config.get_components("services", factory=Service)

        self.log.info("Services:")
        for service in self.services.values():
            self.log.info("  - %s (%s:%s)", service.label, service.host, service.port)

    def list(self):
        """List services"""

        results = {}

        for name, service in self.services.items():
            results[name] = service.info()

        return results 

    def lookup(self, name, key):
        """Lookup data on service component"""

        if name in self.services:
            return getattr(self.services[name], key)

        raise KeyError(f"Unknown service: {name}")


def main():
    DirectoryService(sys.argv).run()



#!/usr/bin/env python
"""Test Inherited init()"""

##############################################################
#
#   Test init() call chain in inherited classes
#
#   2023-07-03  Todd Valentic
#               Initial implementation.
#
##############################################################

import sys

from datatransport import Root, ProcessClient


class MixinA(Root):
    """Mixin class A"""

    def init(self):
        super().init()
        self.log.info("MixinA.init()")


class MixinB(Root):
    """Mixin class B"""

    def init(self):
        super().init()
        self.log.info("MixinB.init()")


class Demo(MixinA, MixinB, ProcessClient):
    """Parent Class"""

    def __init__(self, args, **_kwargs):
        ProcessClient.__init__(self, args)
        MixinA.__init__(self)
        MixinB.__init__(self)

        self.log.info("Demo.mro(): %s", Demo.mro())

    def init(self):
        super().init()

        self.log.info("Demo.init()")


if __name__ == "__main__":
    Demo(sys.argv).run()



#!/usr/bin/env python
"""Console Command"""

###########################################################################
#
#   XML-RPC Service Command Console
#
#   Run commands from an XML-RPC service.
#
#   2006-02-20  Todd Valentic
#               Initial implementation.
#
#   2007-03-11  Todd Valentic
#               Added try..except around connect in do_run()
#
#   2022-10-25  Todd Valentic
#               Python3 port
#               optparse -> argparse
#
#   2023-09-06  Todd Valentic
#               list() return format is now a dictionary
#
#   2023-19-07  Todd Valentic
#               Add pprint option
#
###########################################################################

import argparse
import fnmatch
import cmd
import os
import pprint
import socket
import xmlrpc.client

socket.setdefaulttimeout(100)

VERSION=1.1

class Console(cmd.Cmd):
    """XMLRPC Console"""

    def __init__(self, args):
        cmd.Cmd.__init__(self)

        self.intro = self.make_intro()
        self.prompt = "[not connected] >>> "
        self.options = args
        self.pprint = True

        self.configdir = os.path.expanduser("~/.acorn")
        self.configfile = os.path.join(self.configdir, "history")

        if not os.path.exists(self.configdir):
            os.makedirs(self.configdir)

        self.load_history()

        self.service_cache = {}

    def connect(self, service):
        """Connect to directory"""

        url = self.directory.get(service, "url")
        return xmlrpc.client.ServerProxy(url)

    def make_intro(self):
        """Banner"""

        width = 75
        intro = []
        intro.append("")
        intro.append("-" * width)
        intro.append(f"XML-RPC Service Control Console {VERSION}".center(width))
        intro.append('Type "help" for a list commands'.center(width))
        intro.append("-" * width)
        intro.append("")

        return "\n".join(intro)

    # -- History --------------------------------------------------------

    def load_history(self):
        """Load history"""

        self.history_cache = {}
        self.history_index = 0

        if not os.path.exists(self.configfile):
            return

        with open(self.configfile, "rt", encoding="utf-8") as f:
            for entry in f:
                index, command = entry.split(":", 1)
                index = int(index)
                self.history_cache[index] = command.strip()
                self.history_index = max(self.history_index, index)

        self.trim_history()

    def save_history(self):
        """Save history"""

        with open(self.configfile, "wt", encoding="utf-8") as output:
            keys = sorted(self.history_cache)

            for index in keys:
                output.write(f"{index}: {self.history_cache[index]}")

    def trim_history(self):
        """Trim history entries"""

        cutoff = self.history_index - self.options.maxhistory + 1

        for index in list(self.history_cache.keys()):
            if index < cutoff:
                del self.history_cache[index]

    def do_clear(self, _arg):
        """clear history cache"""

        self.history_cache = {}
        self.history_index = 0
        self.save_history()

    def do_shell(self, args):
        """Run previous commands"""

        args = [int(x) for x in args.split()]
        for index in args:
            if self.onecmd(self.history_cache[index]):
                return True

        return False

    def do_history(self, _args):
        """history - print a list of commands that have been entered"""

        for index in sorted(self.history_cache):
            print(f"{index:2}: {self.history_cache[index]}")

        return False

    # -- Connection ------------------------------------------------------

    def do_connect(self, arg):
        """connect [directory [port]] -- connect to directory service"""

        arg = arg.split()

        if not arg:
            host, port = self.options.host, self.options.port
        elif len(arg) == 1:
            host, port = arg[0], self.options.port
        else:
            host, port = arg

        url = f"http://{host}:{port}"
        self.directory = xmlrpc.client.ServerProxy(url)

        try:
            self.directory.ident()
        except (ConnectionRefusedError, xmlrpc.client.Error) as err:
            print(f"Error connecting to directory on {host}:{port}: {err}")
            return ""

        print(f"Connected to directory at {host}:{port}")

        self.prompt = f"[{host}:{port}] >>> "

        return self.onecmd("refresh")

    def do_refresh(self, _arg):
        """refresh -- Refresh list of services for autocompletion"""

        print("Refreshing service list ...")
        self.service_cache = {}

        try:
            services = self.directory.list()
            if isinstance(services, list):
                # Handle older format 
                for service in services: 
                    label = self.directory.get(service, "label")
                    host = self.directory.get(service, "host")
                    port = self.directory.get(service, "port")
                    self.service_cache[service] = (label, host, port)
            else:
                for service, info in services.items():
                    label = info["label"]
                    host = info["host"]
                    port = info["port"]
                    self.service_cache[service] = (label, host, port)
        except (ConnectionRefusedError, xmlrpc.client.Error) as err:
            print(f"Error connecting to server: {err}")

        print(f"Refreshed -- {len(self.service_cache)} services")

        return False

    # -- Listing ------------------------------------------------------

    def do_services(self, _arg):
        """services -- list registered services"""

        # If we do not have a cache, refresh first

        if not self.service_cache:
            self.onecmd("refresh")

        if not self.service_cache:
            print("No services are registered")
            return False

        names = sorted(self.service_cache)

        print(f"There are {len(names)} services registered:")

        width = max(len(x) for x in names)

        for name in names:
            label, host, port = self.service_cache[name]
            print(f"[{host}:{port}] {name.ljust(width)} - {label}")

        return False

    def do_list(self, arg):
        """list <service> -- list commands in a service"""

        if not arg:
            return self.onecmd("services")

        if arg not in self.service_cache:
            print("no service by that name. need to refresh?")
            return False

        try:
            service = self.connect(arg)
            commands = sorted(service.system.listMethods())
        except (ConnectionRefusedError, xmlrpc.client.Error) as err:
            print(f"problem communicating with service: {err}")
            return False

        print(f"There are {len(commands)} commands in {arg}:")

        for name in commands:
            print(f"  {name}")

        return False

    def complete_list(self, text, line, _begidx, _endidx):
        """Autocompletion handler"""

        names = self.service_cache.keys()
        prefix = line.split()[1].strip()
        matches = fnmatch.filter(names, prefix + "*")
        matches = [name.replace(prefix, text) for name in matches]

        return matches

    # -- Execute ---------------------------------------------------------

    def do_run(self, arg):
        """Run command handler"""

        arg = arg.split()

        if len(arg) < 2:
            print("Need to specify service and method names")
            return False

        try:
            service = self.connect(arg[0])
        except (ConnectionRefusedError, xmlrpc.client.Error) as err:
            print(f"Failed to connect to {arg[0]}: {err}")
            return False

        method = arg[1]
        params = arg[2:]

        func = getattr(service, method)

        try:
            output = func(*params)
        except (ConnectionRefusedError, xmlrpc.client.Error) as err:
            print(f"Problem running command: {err}")
            return False

        if self.pprint:
            pprint.pprint(output)
        else:
            print(output)

        return False

    # -- Options ---------------------------------------------------------

    def do_pprint(self, state):
        """Set pretty printing state (on | off)"""

        if state == "on":
            self.pprint = True 
        elif state == "off":
            self.pprint = False 
        else:
            print(f"pprint is {'on' if self.pprint else 'off'}")

        return False 

    # -- Control ---------------------------------------------------------

    def do_quit(self, _arg):
        """quit -- terminates the application"""
        return True

    def preloop(self):
        cmd.Cmd.preloop(self)
        self.service_cache = {}

    def postloop(self):
        cmd.Cmd.postloop(self)
        self.save_history()
        print("Exiting...")

    def precmd(self, line):
        if line and not line.startswith("!"):
            self.history_index += 1
            self.history_cache[self.history_index] = line.strip()
            self.trim_history()
        return line

    def emptyline(self):
        pass

    # shortcuts

    do_EOF = do_quit
    do_q = do_quit
    do_exit = do_quit
    do_h = do_history

    complete_run = complete_list


def main():
    """Script entry point"""

    parser = argparse.ArgumentParser(
        description="Data Transport XMLRPC Service Console"
    )

    parser.add_argument(
        "-p",
        "--port",
        default=8411,
        type=int,
        help="port (default %(default)s)",
    )

    parser.add_argument(
        "-r",
        "--host",
        default="localhost",
        help="direcotry host (default %(default)s)",
    )

    parser.add_argument(
        "-d",
        "--maxhistory",
        default=100,
        type=int,
        metavar="COUNT",
        help="maximum history depth (default %(default)s",
    )

    args = parser.parse_args()

    try:
        Console(args).cmdloop()
    except KeyboardInterrupt:
        print("")

    print("Exiting")



#!/usr/bin/env python
"""Watchdog"""

############################################################################
#
#   watchdog
#
#   This script monitors the process group clients and reports any that
#   were started but have seemed to stop running (they are registered but
#   don't show up as an active process). In this case, the client is
#   restarted. Messages are also sent to the watchdog list in this case.
#
#   History:
#
#   2000-04-24  TAV
#               Initial implementation
#
#   2002-01-20  Todd Valentic
#               Improved wording of notification message.
#
#   2002-01-24  Todd Valentic
#               Added timestamp to message body.
#
#   2002-01-25  Todd Valentic
#               Added check for self.running
#
#   2002-05-19  Todd Valentic
#               Fixed call to restart client (now include arg string).
#
#   2002-08-28  Todd Valentic
#               sri.transport -> Transport
#               python -> python2
#
#   2003-04-15  Todd Valentic
#               Changed config parameter pollrate -> rate
#               Rate now specified in days/hours/mins/secs
#
#   2004-02-08  Todd Valentic
#               Use DeltaTime for rate
#               Use string methods
#
#   2004-08-08  Todd Valentic
#               Updated to new XML-RPC server interface
#
#   2004-08-25  Todd Valentic
#               Converted to directly reading /proc instead of using ps
#
#   2004-12-30  Todd Valentic
#               Convert from mx.DateTime to datetime
#
#   2022-10-06  Todd Valentic
#               Updated for transport3/python3
#   
#   2023-07-26  Todd Valentic
#               Use config getters
#
############################################################################

import pathlib
import sys

from datatransport import ProcessClient
from datatransport import NewsPoster


class Watchdog(ProcessClient):
    """Watchdog Process Client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

    def init(self):
        super().init()

        self.news_poster = NewsPoster(self)
        self.rate = self.config.get_rate("rate", "5m")

    def get_client_pids(self):
        """Query server for expected client PIDs"""

        pids = {}

        for group in self.server.listgroups():
            for client, info in self.server.listclients(group).items():
                pid = info[1]
                if pid:
                    pids[pid] = (group, client)

        return pids

    def get_running_pids(self):
        """Get list of running PIDs on system"""

        proc = pathlib.Path("/proc")
        return [int(x.name) for x in proc.iterdir() if x.name.isdigit()]

    def post_message(self, pid, group_name, client_name):
        """Post a trouble message"""

        body = []
        body.append("")
        body.append(
            "  The watchdog has detected the following client has stopped running:"
        )
        body.append("")
        body.append(f"      Client name: {client_name}")
        body.append(f"       Group name: {group_name}")
        body.append(f"       Process ID: {pid}")
        body.append(f"       Time stamp: {self.now()}")
        body.append("")
        body.append("  The client has been restarted.")
        body.append("")

        header = f"Watchdog restart {client_name}"

        try:
            self.news_poster.set_subject(header)
            self.news_poster.post_text("\n".join(body))
        except:  # pylint: disable=bare-except
            self.log.error("Error posting notification message to the news server.")

    def restart(self, pid, group, client):
        """Restart the client process"""

        self.log.info("Process '%s %s' (PID %s) is missing", group, client, pid)
        self.log.info("  Attempting to restart")

        self.post_message(pid, group, client)

        try:
            self.server.startclient(group, client)
        except:  # pylint: disable=bare-except
            self.log.info("  Error detected in client restart!")

    def checkup(self):
        """Check to see if any clients have unexpectedly stopped"""

        self.log.info("Checking processes")

        try:
            self.server.status()
        except:  # pylint: disable=bare-except
            self.log.error("Cannot connect to the transport server!")
            return

        client_pids = self.get_client_pids()
        running_pids = self.get_running_pids()

        for pid, entry in client_pids.items():
            group, client = entry

            if not pid in running_pids:
                self.restart(pid, group, client)

    def main(self):
        """Main application"""

        while self.wait(self.rate):
            self.checkup()


if __name__ == "__main__":
    Watchdog(sys.argv).run()



#!/usr/bin/env python3
"""Remove a newsgroup"""

#####################################################################
#
#   Remove a news group
#
#   2016-12-26  Todd Valentic
#               Use datatransport package
#
#   2023-08-23  Todd Valentic
#               Updated for transport3 / python3
#               Add command line options
#               Allow for multiple newsgroups
#               Allow for patterns
#
#####################################################################

import argparse
import fnmatch
import sys

from datatransport import newstool

VERSION = "1.0"


def main():
    """Script entry point"""

    desc = "Remove a newsgroup from the news server"
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument("-V", "--version", action="version", version=VERSION)

    parser.add_argument(
        "-s", "--server", default="localhost", help="News server (default: localhost)"
    )
    parser.add_argument(
        "-p", "--port", default=119, type=int, help="News server port (default: 119)"
    )
    parser.add_argument(
        "-y", "--yes", action="store_true", help="Automatically confirm removal"
    )

    parser.add_argument("newsgroups", nargs="+", metavar="newsgroup")

    args = parser.parse_args()

    server = newstool.NewsControl()
    server.set_server(args.server, port=args.port)

    available_newsgroups = server.list_newsgroups()

    remove_newsgroups = set()

    for pattern in args.newsgroups:
        for newsgroup in available_newsgroups:
            if fnmatch.fnmatch(newsgroup, pattern):
                remove_newsgroups.add(newsgroup)

    if not remove_newsgroups:
        print("No newsgroups were found")
        sys.exit(0)

    if not args.yes:
        print("")
        print(f"Newsgroups to be removed on {args.server}:{args.port}:")

        for newsgroup in remove_newsgroups:
            print(f"  {newsgroup}")
        print("")

        reply = input("Confirm removal? (Y/N) ").lower()
        if reply != "y":
            sys.exit(1)

    for newsgroup in remove_newsgroups:
        print(f"Removing {newsgroup}")
        server.rmgroup(newsgroup)

    sys.exit(0)



#!/usr/bin/env python3
"""RealTimeFeed Component"""

###########################################################################
#
#   RealTimeFeed
#
#   This program acts as a relay between news groups on two different news
#   servers. The primary intent was to quickly deliver messages outside of
#   the normal hourly cron job. A message is extracted from a newsgroup,
#   it's header stripped of fields that will be reset when it is sent, and
#   then it is posted to the newsgroup on the out going server. The news
#   groups do not need to be the same.
#
#   1.0.0   2000-01-27  TAV
#           Initial implementation
#
#   1.0.1   2002-01-25  Todd Valentic
#           Updated to new NewsPollMixin interface.
#           Removed config param "realtime.server" - just use the
#               (now) standard post.newsserver parameter.
#
#   1.0.2   2002-08-27  Todd Valentic
#           Setup with configure script.
#           sri.transport -> Transport
#
#   1.0.3   2003-03-08  Todd Valentic
#           Added try..except around post to protect against errors.
#
#   1.0.4   2003-06-12  Todd Valentic
#           Fixed a bug: NewsPoster's groupHeader is now newsgroupHeader.
#
#   1.0.5   2003-06-30  Todd Valentic
#           Updated to new NewsPoller interface.
#
#   1.0.6   2003-09-09  Todd Valentic
#           When converting a message into a string, use the
#               .as_string() method. Just casting it as str()
#               causes the unix from envolope header to be included.
#
#   1.0.7   2006-10-05  Todd Valentic
#           Added validation.
#
#   1.0.8   2009-05-11  Todd Valentic
#           NewsPoster.postRaw() now takes an email.Message object.
#
#   1.09    2009-09-14  Todd Valentic
#           Raise ProcessRetry if we fail to post.
#
#   1.10    2011-06-28  Todd Valentic
#           Use a copy of the message when posting. Found a bug when
#               the destination group name is different, but built
#               from the name in the message. A process retry will
#               send back in the modified messages, leading to strange
#               errors.
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#
#   2023-07-04  Todd Valentic
#               Updates for transport3 / python3
#                   NewsPoller, NewsPoster
#
###########################################################################

import copy
import nntplib
import sys

from datatransport import ProcessClient
from datatransport import newstool
from datatransport import NewsPoster
from datatransport import NewsPoller


class RealTimeFeed(ProcessClient):
    """RealTimeFeed Process Client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.news_poster = NewsPoster(self)
        self.news_poller = NewsPoller(self, callback=self.process)
        self.main = self.news_poller.main

    def filter_headers(self, message):
        """Filter out headers"""

        ignore_headers = [
            "Newsgroups",
            "Path",
            "Date",
            "Injection-Date",
            "Injection-Info",
            "Message-ID",
            "NNTP-Posting-Host",
            "NNTP-Posting-Date",
            "X-Trace",
            "X-Complaints-To",
            "Xref",
        ]

        for header in ignore_headers:
            if header in message:
                del message[header]

        message["Newsgroups"] = self.news_poster.newsgroup_header

        return message

    def valid_message(self, _message):
        """Validate message"""
        return True

    def process(self, src_message):
        """Process new message"""

        message = copy.deepcopy(src_message)

        if not self.valid_message(message):
            self.log.debug("Skipping message (not valid)")
            return

        message = self.filter_headers(message)

        try:
            response = self.news_poster.post_raw(message)
        except nntplib.NNTPError as err:
            self.log.error("Problem posting to the news server: %s", err)
            raise newstool.ProcessRetry

        self.log.info("Files posted, response=%s", response)

def main():
    """Script entry point"""
    RealTimeFeed(sys.argv).run()



#!/usr/bin/env python
"""List running transport processes"""

##########################################################################
#
#   List running transport processes
#
#   2020-10-26  Todd Valentic
#               Initial implementation. Based on original transportps
#
##########################################################################

from pathlib import Path


def find_server(name):
    """Find data transport server process ID"""
    for pid in Path("/proc").iterdir():
        if pid.name.isdigit():
            comm = pid.joinpath("comm").read_text().strip()
            if comm == name:
                return pid

    return None


def list_processes():
    """Show data transport processes"""

    parent = find_server("transportd")

    if parent is None:
        print("No transport processes found")
        return

    ppid = parent.name

    children = parent.joinpath("task", ppid, "children").read_text()

    pids = [ppid]
    pids.extend(children.split())

    for pid in pids:
        path = Path(f"/proc/{pid}/cmdline")
        cmdline = path.read_bytes().replace(b"\0", b" ").decode()
        _interpreter, cmd, args = cmdline.split(" ", 2)
        cmd = Path(cmd).name
        print(f"[{pid:>7}] {cmd} {args}")


def main():
    """Main application"""

    print("-" * 75)
    list_processes()
    print("-" * 75)



#!/usr/bin/env python

from Transport  import ProcessClient
from datetime   import datetime 

import sys
import os

class CreateFiles(ProcessClient):
	
	def __init__(self,argv):
		ProcessClient.__init__(self,argv)

		self.pollrate	= self.getDeltaTime('pollrate',60)
		self.path		= self.get('filename.path','.')
		self.filenames	= self.get('filenames','%Y%m%d-%H%M%S.txt').split()

	def run(self):

		while self.running:

			curdate	= datetime.now()

			for filename in self.filenames:
				filename = curdate.strftime(filename)
				filename = os.path.join(self.path,filename)

				dir = os.path.dirname(filename)
				if not os.path.exists(dir):
					os.makedirs(dir)

				file = open(filename,'w')
				file.write('%s\n' % datetime.now().strftime('%Y%m%d %H%M%S'))
				file.close()

				self.log.info('Created: %s' % filename)

			self.wait(self.pollrate)

if __name__ == '__main__':
	CreateFiles(sys.argv).run()
	
